"""
.. _reading_and_exploring_NWB_File:

Reading and Exploring an NWB File
==================================

This tutorial will demonstrate how to read, explore, and do basic visualizations with
an NWB File using different tools.

An :py:class:`~pynwb.file.NWBFile` represents a single session of an experiment.
It contains all the data of that session and the metadata required to understand the data.

We will demonstrate how to use the `DANDI <https://gui.dandiarchive.org/>`_ neurophysiology data archive to access
the data in two different ways: (1) by downloading it to your computer and (2) streaming it.

We will briefly show tools for exploring NWB Files interactively and refer the reader to the
:nwb_overview:`NWB Overview <tools/analysis_tools_home.html>` documentation for more details about the available tools.


.. seealso::
    You can learn more about the :py:class:`~pynwb.file.NWBFile` format in the :ref:`basics` tutorial.


The following examples will reference variables that may not be defined within the block they are used in. For
clarity, we define them here:
"""

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_read_basics.png'

import matplotlib.pyplot as plt
import numpy as np

from pynwb import NWBHDF5IO

####################
# We will access NWB data on the `DANDI Archive <https://gui.dandiarchive.org/>`_,
# and demonstrate reading one session of an experiment by
# `Chandravadia et al. (2020) <https://www.nature.com/articles/s41597-020-0415-9>`_. In this study,
# the authors recorded single neuron activity from the medial temporal lobes of human subjects
# while they performed a recognition memory task.
#
# Download the data
# -----------------
# First, we will demonstrate how to download an NWB data file from `DANDI <https://gui.dandiarchive.org/>`_
# to your machine.
#
# Download using the DANDI Web UI
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# You can download files directly from the DANDI website.
#
# 1. Go to the DANDI archive and open `this <https://gui.dandiarchive.org/dandiset/000004/draft>`_ dataset
# 2. List the files in this dataset by clicking the "Files" button in Dandiset Actions (top right column of the page).
#
# .. image:: ../../_static/demo_dandi_view_files_in_dataset.png
#   :width: 700
#   :alt: view files on dandi
#   :align: center
#
# 3. Choose the folder "sub-P11MHM" by clicking on its name.
#
# .. image:: ../../_static/demo_dandi_select_folder.png
#   :width: 700
#   :alt: selecting a folder on dandi
#   :align: center
#
# 4. Download the NWB data file "sub-P11HMH_ses-20061101_ecephys+image.nwb" to your
# computer by clicking on the download symbol.
#
# .. image:: ../../_static/demo_dandi_download_data.png
#   :width: 700
#   :alt: selecting a folder on dandi
#   :align: center
#
# Downloading data programmatically
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Alternatively, you can download data using the `dandi` Python module.

from dandi.download import download

download("https://api.dandiarchive.org/api/assets/0f57f0b0-f021-42bb-8eaa-56cd482e2a29/download/", ".")

######################################################
# .. seealso::
#
#   Learn about all the different ways you can download data from the DANDI Archive
#   `here <https://www.dandiarchive.org/handbook/12_download/>`_
#
# .. seealso:: Streaming data
#
#   Instead of downloading data, another approach is to stream data directly from an archive. Streaming data allows you
#   to download only the data you want from a file, so it can be a much better approach when the desired data files
#   contain a lot of data you don't care about. There are several approaches to streaming NWB files, outlined in
#   :ref:`streaming`.
#
# Opening an NWB file with NWBHDF5IO
# ----------------------------------
#
# Reading and writing NWB data is carried out using the :py:class:`~pynwb.NWBHDF5IO` class.
# :py:class:`~pynwb.NWBHDF5IO` reads NWB data that is in the `HDF5 <https://www.hdfgroup.org/solutions/hdf5/>`_
# storage format, a popular, hierarchical format for storing large-scale scientific data.
#
# The first argument to the constructor of :py:class:`~pynwb.NWBHDF5IO` is the ``file_path``. Use the ``read`` method to
# read the data into a :py:class:`~pynwb.file.NWBFile` object.

filepath = "sub-P11HMH_ses-20061101_ecephys+image.nwb"
# Open the file in read mode "r",
io = NWBHDF5IO(filepath, mode="r")
nwbfile = io.read()
nwbfile

#######################################
# :py:class:`~pynwb.NWBHDF5IO` can also be used as a context manager:

with NWBHDF5IO(filepath, mode="r") as io2:
    nwbfile2 = io2.read()

    # data accessible here

# data not accessible here

######################################
# The advantage of using a context manager is that the file is closed automatically when the context finishes
# successfully or if there is an error. Be aware that if you use this method, closing the context (unindenting the code)
# will automatically close the :py:class:`~pynwb.NWBHDF5IO` object and the corresponding h5py File object. The data not
# already read from the NWB file will then be inaccessible, so any code that reads data must be placed within the
# context.
#
# Access stimulus data
# --------------------
#
# Data representing stimuli that were presented to the experimental subject are stored in
# :py:class:`~pynwb.file.NWBFile.stimulus` within the :py:class:`~pynwb.file.NWBFile` object.

nwbfile.stimulus

####################
# ``NWBFile.stimulus`` is a dictionary that can contain PyNWB objects representing
# different types of data, such as images (grayscale, RGB) or time series of images.
# In this file, ``NWBFile.stimulus`` contains a single key "StimulusPresentation" with an
# :py:class:`~pynwb.image.OpticalSeries` object representing what images were shown to the subject and at what times.

nwbfile.stimulus["StimulusPresentation"]

####################
# Lazy loading of datasets
# ------------------------
# Data arrays are read passively from the NWB file.
# Accessing the ``data`` attribute of the :py:class:`~pynwb.image.OpticalSeries` object
# does not read the data values, but presents an :py:class:`h5py.Dataset` object that can be indexed to read data.
# You can use the ``[:]`` operator to read the entire data array into memory.

stimulus_presentation = nwbfile.stimulus["StimulusPresentation"]
all_stimulus_data = stimulus_presentation.data[:]

####################
# Images may be 3D or 4D (grayscale or RGB), where the first dimension must be time (frame).
# The second and third dimensions represent x and y.
# The fourth dimension represents the RGB value (length of 3) for color images.

stimulus_presentation.data.shape

####################
# This :py:class:`~pynwb.image.OpticalSeries` data contains 200 images of size 400x300 pixels with three channels
# (red, green, and blue).
#
# Slicing datasets
# ----------------
# It is often preferable to read only a portion of the data.
# To do this, index or slice into the ``data`` attribute just like if you were
# indexing or slicing a numpy array.

frame_index = 31
image = stimulus_presentation.data[frame_index]
# Reverse the last dimension because the data were stored in BGR instead of RGB
image = image[..., ::-1]
plt.imshow(image, aspect="auto")

####################
# Access single unit data
# -----------------------
# Data and metadata about sorted single units are stored in :py:class:`~pynwb.misc.Units`
# object. It stores metadata about each single unit in a tabular form, where each row represents
# a unit with spike times and additional metadata.
#
# .. seealso::
#     You can learn more about units in the :ref:`ecephys_tutorial` tutorial.
#

units = nwbfile.units

####################
# We can view the single unit data as a :py:class:`~pandas.DataFrame`.

units_df = units.to_dataframe()
units_df.head()

####################
# To access the spike times of the first single unit, index :py:class:`~pynwb.file.NWBFile.units` with the column
# name "spike_times" and then the row index, 0. All times in NWB are stored in seconds
# relative to the session start time.

units["spike_times"][0]

####################
# Visualize spiking activity relative to stimulus onset
# -----------------------------------------------------
# We can look at when these single units spike relative to when image stimuli were presented to the subject.
# We will iterate over the first 3 units and get their spike times.
# Then for each unit, we will iterate over each stimulus onset time and compute the spike times relative
# to stimulus onset. Finally, we will create a raster plot and histogram of these aligned spike times.

before = 1.0  # in seconds
after = 3.0

# Get the stimulus times for all stimuli
# get_timestamps() works whether the time is stored as an array of timestamps or as
# starting time and sampling rate.
stim_on_times = stimulus_presentation.get_timestamps()

for unit in range(3):
    unit_spike_times = nwbfile.units["spike_times"][unit]
    trial_spikes = []
    for time in stim_on_times:
        # Compute spike times relative to stimulus onset
        aligned_spikes = unit_spike_times - time
        # Keep only spike times in a given time window around the stimulus onset
        aligned_spikes = aligned_spikes[
            (-before < aligned_spikes) & (aligned_spikes < after)
        ]
        trial_spikes.append(aligned_spikes)
    fig, axs = plt.subplots(2, 1, sharex="all")
    plt.xlabel("time (s)")
    axs[0].eventplot(trial_spikes)

    axs[0].set_ylabel("trial")
    axs[0].set_title("unit {}".format(unit))
    axs[0].axvline(0, color=[0.5, 0.5, 0.5])

    axs[1].hist(np.hstack(trial_spikes), 30)
    axs[1].axvline(0, color=[0.5, 0.5, 0.5])

####################
# Access Trials
# -------------
# Trials are stored as :py:class:`~pynwb.epoch.TimeIntervals` object which is a subclass
# of :py:class:`~hdmf.common.table.DynamicTable`. :py:class:`~hdmf.common.table.DynamicTable` objects are used to store
# metadata about each trial in a tabular form, where each row represents a trial and has a start time, stop time,
# and additional metadata.
#
# .. seealso::
#     You can learn more about trials in the :ref:`time_intervals` tutorial.
#
# Similarly to :py:class:`~pynwb.misc.Units`, we can view trials as a :py:class:`pandas.DataFrame`.

trials_df = nwbfile.trials.to_dataframe()
trials_df.head()

####################
# The stimulus can be mapped one-to-one to each row (trial) of
# :py:class:`~pynwb.file.NWBFile.trials` based on the ``stim_on_time`` column.

assert np.all(stimulus_presentation.timestamps[:] == trials_df.stim_on_time[:])

####################
# Visualize the first 3 images that were categorized as landscapes in the session:

stim_on_times_landscapes = trials_df[
    trials_df.category_name == "landscapes"
].stim_on_time
for time in stim_on_times_landscapes.iloc[:3]:
    img = np.squeeze(
        stimulus_presentation.data[
            np.where(stimulus_presentation.timestamps[:] == time)
        ]
    )
    # Reverse the last dimension because the data were stored in BGR instead of RGB
    img = img[..., ::-1]
    plt.figure()
    plt.imshow(img, aspect="auto")

####################
# Exploring the NWB file
# ----------------------
# So far we have explored the NWB file by printing the :py:class:`~pynwb.file.NWBFile`
# object and accessing its attributes, but it may be useful to explore the data in a
# more interactive, visual way. See :ref:`analysistools-explore` for an updated list of programs for
# exploring NWB files.

####################
# Close the open NWB file
# -----------------------
# It is good practice, especially on Windows, to close any files that you have opened.

io.close()



from hdmf.common import HERD as hdmf_HERD
from . import get_type_map as tm
from hdmf.utils import docval, get_docval


class HERD(hdmf_HERD):
    """
    HDMF External Resources Data Structure.
    A table for mapping user terms (i.e. keys) to resource entities.
    """
    @docval(*get_docval(hdmf_HERD.__init__))
    def __init__(self, **kwargs):
        kwargs['type_map'] = tm()
        super().__init__(**kwargs)



"""
How to Configure Term Validations
=================================

This is a user guide for how to curate and take advantage of configuration files in
order to more easily validate terms within datasets or attributes.

Introduction
-------------
Users will create a configuration YAML file that outlines the fields (within a neurodata type)
they want to be validated against a set of allowed terms. 
After creating the configuration file, users will need to load the
configuration file with the :py:func:`~pynwb.load_type_config` method.
With the configuration loaded, every instance of the neurodata
types defined in the configuration file will have the respective fields wrapped with a
:py:class:`~hdmf.term_set.TermSetWrapper`.
This automatic wrapping is what provides the term validation for the field value.
For greater control on which datasets and attributes are validated
against which sets of allowed terms, use the 
:py:class:`~hdmf.term_set.TermSetWrapper` on individual datasets and attributes instead.
You can follow the 
`TermSet tutorial in the HDMF documentation 
<https://hdmf.readthedocs.io/en/stable/tutorials/plot_term_set.html#sphx-glr-tutorials-plot-term-set-py>`_
for more information.

To unload a configuration, simply call :py:func:`~pynwb.unload_type_config`.
We also provide a helper method to see the configuration that has been loaded:
:py:func:`~pynwb.get_loaded_type_config`


How to make a Configuration File
--------------------------------
To see an example of a configuration file, please refer to
`<https://github.com/NeurodataWithoutBorders/pynwb/tree/dev/src/pynwb/config/nwb_config.yaml>`_.
The configuration file uses YAML syntax. The
user will construct a series of nested dictionaries to encompass all the necessary information.

1. The user needs to define all the relevant namespaces. Recall that each neurodata type exists within
   a namespace, whether that is the core namespace in PyNWB or a namespace in an extension. As namespaces grow,
   we also require a version to be recorded in the configuration file to ensure proper functionality.
2. Within a namespace dictionary, the user will have a list of data types they want to configure.
3. Each data type will have a list of fields associated with a :py:class:`~hdmf.term_set.TermSet`.
   The user can use the same or unique TermSet instances for each field.
"""
try:
    import linkml_runtime  # noqa: F401
except ImportError as e:
    raise ImportError("Please install linkml-runtime to run this example: pip install linkml-runtime") from e

from dateutil import tz
from datetime import datetime
from uuid import uuid4
import os

from pynwb import NWBFile, get_loaded_type_config, load_type_config, unload_type_config
from pynwb.file import Subject

####################################
# How to use a Configuration file
# -------------------------------
# As mentioned prior, the first step after creating a configuration file is
# to load the file. In this configuration file, we have defined two fields
# we want to always be validated: ``experimenter`` and ``species``. Each of these
# are from a different neurodata type, :py:class:`~pynwb.file.NWBFile` and
# :py:class:`~pynwb.file.Subject` respectively, and each
# have a unique associated :py:class:`~hdmf.term_set.TermSet`.
# It is important to remember that with the configuration loaded, the fields
# are wrapped automatically, meaning the user should proceed with creating
# the instances normally, i.e., without wrapping directly. Once instantiated,
# the value of the fields are wrapped and then validated to see if it is a
# permissible value in their respective :py:class:`~hdmf.term_set.TermSet`.

try:
    dir_path = os.path.dirname(os.path.abspath(__file__))  # when running as a .py
except NameError:
    dir_path = os.path.dirname(os.path.abspath("__file__"))  # when running as a script or notebook
yaml_file = os.path.join(dir_path, 'nwb_gallery_config.yaml')
load_type_config(config_path=yaml_file)

session_start_time = datetime(2018, 4, 25, hour=2, minute=30, second=3, tzinfo=tz.gettz("US/Pacific"))

nwbfile = NWBFile(
    session_description="Mouse exploring an open field",  # required
    identifier=str(uuid4()),  # required
    session_start_time=session_start_time,  # required
    session_id="session_1234",  # optional
    experimenter=[
        "Bilbo Baggins",
    ],  # optional
    lab="Bag End Laboratory",  # optional
    institution="University of My Institution",  # optional
    experiment_description="I went on an adventure to reclaim vast treasures.",  # optional
    related_publications="DOI:10.1016/j.neuron.2016.12.011",  # optional
)

subject = Subject(
    subject_id="001",
    age="P90D",
    description="mouse 5",
    species="Mus musculus",
    sex="M",
)

nwbfile.subject = subject

####################################
# How to see the Configuration file
# ---------------------------------
# Call :py:class:`~pynwb.get_loaded_type_config` to get a dictionary containing the
# current configuration.
config = get_loaded_type_config()

######################################
# How to unload the Configuration file
# ------------------------------------
# Call :py:class:`~pynwb.unload_type_config` to toggle off the automatic validation.
unload_type_config()



from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np
from hdmf.utils import docval

from pynwb import NWBFile, TimeSeries, available_namespaces
from pynwb.core import NWBContainer, NWBData
from pynwb.testing import TestCase


class MyTestClass(NWBContainer):

    __nwbfields__ = ("prop1", "prop2")

    @docval({"name": "name", "type": str, "doc": "The name of this container"})
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop1 = "test1"


class TestNWBContainer(TestCase):
    def test_constructor(self):
        """Test constructor
        """
        obj = MyTestClass("obj1")
        self.assertEqual(obj.name, "obj1")
        obj.prop2 = "test2"

    def test_nwbfields(self):
        """Test that getters and setters work for nwbfields
        """
        obj = MyTestClass("obj1")
        obj.prop2 = "test2"
        self.assertEqual(obj.prop1, "test1")
        self.assertEqual(obj.prop2, "test2")

    def test_get_data_type(self):
        obj = NWBContainer("obj1")
        dt = obj.data_type
        self.assertEqual(dt, 'NWBContainer')


class MyNWBData(NWBData):

    __nwbfields__ = ("data", )

    @docval(
        {"name": "name", "type": str, "doc": "The name of this container"},
        {"name": "data", "type": ("array_data", "data"), "doc": "any data"},
    )
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class TestNWBData(TestCase):
    def test_constructor(self):
        """Test constructor
        """
        obj = MyNWBData("obj1", data=[[1, 2, 3], [1, 2, 3]])
        self.assertEqual(obj.name, "obj1")

    def test_append_list(self):

        obj = MyNWBData("obj1", data=[[1, 2, 3], [1, 2, 3]])
        obj.append([4, 5, 6])
        np.testing.assert_array_equal(obj.data, [[1, 2, 3], [1, 2, 3], [4, 5, 6]])

    def test_append_ndarray_2d(self):
        obj = MyNWBData("obj1", data=np.array([[1, 2, 3], [1, 2, 3]]))
        obj.append([4, 5, 6])
        np.testing.assert_array_equal(obj.data, [[1, 2, 3], [1, 2, 3], [4, 5, 6]])

        def test_append_ndarray_1d(self):
            obj = MyNWBData("obj1", data=np.array([1, 2, 3]))
            obj.append([4])
            np.testing.assert_array_equal(obj.data, [1, 2, 3, 4])

    def test_extend_list(self):
        obj = MyNWBData("obj1", data=[[1, 2, 3], [1, 2, 3]])
        obj.extend([[4, 5, 6]])
        np.testing.assert_array_equal(obj.data, [[1, 2, 3], [1, 2, 3], [4, 5, 6]])

    def test_extend_ndarray_1d(self):
        obj = MyNWBData("obj1", data=np.array([1, 2, 3]))
        obj.extend([4, 5, 6])
        np.testing.assert_array_equal(obj.data, [1, 2, 3, 4, 5, 6])

    def test_extend_ndarray_2d(self):
        obj = MyNWBData("obj1", data=np.array([[1, 2, 3], [1, 2, 3]]))
        obj.extend([[4, 5, 6]])
        np.testing.assert_array_equal(obj.data, [[1, 2, 3], [1, 2, 3], [4, 5, 6]])


class TestPrint(TestCase):
    def test_print_file(self):
        nwbfile = NWBFile(
            session_description="session_description",
            identifier="identifier",
            session_start_time=datetime.now(tzlocal()),
        )
        ts1 = TimeSeries(
            name="name1",
            data=[1000, 2000, 3000],
            unit="unit",
            timestamps=[1.0, 2.0, 3.0],
        )
        ts2 = TimeSeries(
            name="name2",
            data=[1000, 2000, 3000],
            unit="unit",
            timestamps=[1.0, 2.0, 3.0],
        )
        expected = """name1 pynwb.base.TimeSeries at 0x%d
Fields:
  comments: no comments
  conversion: 1.0
  data: [1000 2000 3000]
  description: no description
  interval: 1
  offset: 0.0
  resolution: -1.0
  timestamps: [1. 2. 3.]
  timestamps_unit: seconds
  unit: unit
"""
        expected %= id(ts1)
        self.assertEqual(str(ts1), expected)
        nwbfile.add_acquisition(ts1)
        nwbfile.add_acquisition(ts2)
        nwbfile.add_epoch(start_time=1.0, stop_time=10.0, tags=["tag1", "tag2"])
        expected_re = r"""root pynwb\.file\.NWBFile at 0x\d+
Fields:
  acquisition: {
    name1 <class 'pynwb\.base\.TimeSeries'>,
    name2 <class 'pynwb\.base\.TimeSeries'>
  }
  epochs: epochs <class 'pynwb.epoch.TimeIntervals'>
  file_create_date: \[datetime.datetime\(.*\)\]
  identifier: identifier
  session_description: session_description
  session_start_time: .*
  timestamps_reference_time: .*
"""
        self.assertRegex(str(nwbfile), expected_re)


class TestAvailableNamespaces(TestCase):
    def test_available_namespaces(self):
        self.assertEqual(
            available_namespaces(), ("hdmf-common", "hdmf-experimental", "core")
        )



"""
.. _object_ids:

Object IDs in NWB
=================

This example focuses on how to access object IDs from NWB container objects and NWB container objects by
object ID. Every NWB container object has an object ID that is a UUID_ string, such as
"123e4567-e89b-12d3-a456-426655440000". These IDs have a non-zero probability of being duplicated, but are practically
unique and used widely across computing platforms as if they are unique.

The object ID of an NWB container object can be accessed using the
:py:attr:`~hdmf.container.AbstractContainer.object_id` method.

.. _UUID: https://en.wikipedia.org/wiki/Universally_unique_identifier

"""

from datetime import datetime

import numpy as np
from dateutil.tz import tzlocal

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_objectid.png'
from pynwb import NWBFile, TimeSeries

# set up the NWBFile
start_time = datetime(2019, 4, 3, 11, tzinfo=tzlocal())
nwbfile = NWBFile(
    session_description="demonstrate NWB object IDs",
    identifier="NWB456",
    session_start_time=start_time,
)

# make some simulated data
timestamps = np.linspace(0, 100, 1024)
data = (
    np.sin(0.333 * timestamps)
    + np.cos(0.1 * timestamps)
    + np.random.randn(len(timestamps))
)
test_ts = TimeSeries(name="raw_timeseries", data=data, unit="m", timestamps=timestamps)

# add it to the NWBFile
nwbfile.add_acquisition(test_ts)

# print the object ID of the NWB file
print(nwbfile.object_id)

# print the object ID of the TimeSeries
print(test_ts.object_id)

####################
# The :py:class:`~pynwb.file.NWBFile` class has the :py:meth:`~pynwb.file.NWBFile.objects` property, which provides a
# dictionary of all neurodata_type objects in the `NWBFile`, indexed by each object's object ID.

print(nwbfile.objects)

####################
# You can iterate through the `objects` dictionary as with any other Python dictionary.

for oid in nwbfile.objects:
    print(nwbfile.objects[oid])

for obj in nwbfile.objects.values():
    print('%s: %s "%s"' % (obj.object_id, obj.neurodata_type, obj.name))

####################
# If you have stored the object ID of a particular NWB container object, you can use it as a key on `NWBFile.objects` to
# get the object.

ts_id = test_ts.object_id
my_ts = nwbfile.objects[ts_id]  # test_ts == my_ts

####################
#
# .. note::
#    It is important to note that the object ID is NOT a unique hash of the data. If the contents of an NWB container
#    change, the object ID remains the same.
#



"""
Allen Brain Observatory
=================================

Create an nwb file from Allen Brain Observatory data.
"""

########################################
# This example demonstrates the basic functionality of several parts of the pynwb write API, centered around the optical
# physiology submodule (pynwb.ophys). We will use the allensdk as a read API, while leveraging the pynwb data model and
# write api to transform and write the data back to disk.
#
# .. note: Using the latest allensdk package requires Python 3.6 or higher.

########################################
# .. raw:: html
#     :url: https://gist.githubusercontent.com/nicain/82e6b3d8f9ff5b85ef01a582e41e2389/raw/

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_allenbrainobservatory.png'

import allensdk.brain_observatory.stimulus_info as si
from allensdk.core.brain_observatory_cache import BrainObservatoryCache

from pynwb import NWBHDF5IO, NWBFile, TimeSeries
from pynwb.device import Device
from pynwb.image import ImageSeries, IndexSeries
from pynwb.ophys import DfOverF, ImageSegmentation, OpticalChannel

# Settings:
ophys_experiment_id = 562095852
save_file_name = "brain_observatory.nwb"

########################################
# Let's begin by downloading an Allen Institute Brain Observatory file.  After we cache this file locally (approx. 450
# MB), we can open data assets we wish to write into our NWB:N file.  These include stimulus, acquisition, and
# processing data, as well as time "epochs" (intervals of interest)".
boc = BrainObservatoryCache(manifest_file="manifest.json")
dataset = boc.get_ophys_experiment_data(ophys_experiment_id)
metadata = dataset.get_metadata()
cell_specimen_ids = dataset.get_cell_specimen_ids()
timestamps, dFF = dataset.get_dff_traces()
stimulus_list = [
    s for s in si.SESSION_STIMULUS_MAP[metadata["session_type"]] if s != "spontaneous"
]
running_data, _ = dataset.get_running_speed()
trial_table = dataset.get_stimulus_table("master")
trial_table["start"] = timestamps[trial_table["start"].values]
trial_table["end"] = timestamps[trial_table["end"].values]
epoch_table = dataset.get_stimulus_epoch_table()
epoch_table["start"] = timestamps[epoch_table["start"].values]
epoch_table["end"] = timestamps[epoch_table["end"].values]

########################################
# 1) First, lets create a top-level "file" container object.  All the other NWB:N data components will be stored
# hierarchically, relative to this container.  The data won't actually be written to the file system until the end of
# the script.

nwbfile = NWBFile(
    session_description="Allen Brain Observatory dataset",
    identifier=str(metadata["ophys_experiment_id"]),
    session_start_time=metadata["session_start_time"],
)


########################################
# 2) Next, we add stimuli templates (one for each type of stimulus), and a data series that indexes these templates to
# describe what stimulus was being shown during the experiment.
for stimulus in stimulus_list:
    visual_stimulus_images = ImageSeries(
        name=stimulus,
        data=dataset.get_stimulus_template(stimulus),
        unit="NA",
        format="raw",
        timestamps=[0.0],
    )
    image_index = IndexSeries(
        name=stimulus,
        data=dataset.get_stimulus_table(stimulus).frame.values,
        unit="NA",
        indexed_timeseries=visual_stimulus_images,
        timestamps=timestamps[dataset.get_stimulus_table(stimulus).start.values],
    )
    nwbfile.add_stimulus_template(visual_stimulus_images)
    nwbfile.add_stimulus(image_index)

########################################
# 3) Besides the two-photon calcium image stack, the running speed of the animal was also recorded in this experiment.
# We can store this data as a TimeSeries, in the acquisition portion of the file.

running_speed = TimeSeries(
    name="running_speed", data=running_data, timestamps=timestamps, unit="cm/s"
)

nwbfile.add_acquisition(running_speed)

########################################
# 4) In NWB:N, an "epoch" is an interval of experiment time that can slice into a timeseries (for example running_speed,
# the one we just added).  PyNWB uses an object-oriented approach to create links into these timeseries, so that data is
# not copied multiple times.  Here, we extract the stimulus epochs (both fine and coarse-grained) from the Brain
# Observatory experiment using the allensdk.

for _, row in trial_table.iterrows():
    nwbfile.add_epoch(
        start_time=row.start,
        stop_time=row.end,
        timeseries=[running_speed],
        tags="trials",
    )

for _, row in epoch_table.iterrows():
    nwbfile.add_epoch(
        start_time=row.start,
        stop_time=row.end,
        timeseries=[running_speed],
        tags="stimulus",
    )

########################################
# 5) In the brain observatory, a two-photon microscope is used to acquire images of the calcium activity of neurons
# expressing a fluorescent protein indicator.  Essentially the microscope captures picture (30 times a second) at a
# single depth in the visual cortex (the imaging plane).  Let's use pynwb to store the metadata associated with this
# hardware and experimental setup:
optical_channel = OpticalChannel(
    name="optical_channel",
    description="2P Optical Channel",
    emission_lambda=520.0,
)

device = Device(metadata["device"])
nwbfile.add_device(device)

imaging_plane = nwbfile.create_imaging_plane(
    name="imaging_plane",
    optical_channel=optical_channel,
    description="Imaging plane ",
    device=device,
    excitation_lambda=float(metadata["excitation_lambda"].split(" ")[0]),
    imaging_rate=30.0,
    indicator="GCaMP6f",
    location=metadata["targeted_structure"],
    conversion=1.0,
    unit="unknown",
    reference_frame="unknown",
)

########################################
# The Allen Institute does not include the raw imaging signal, as this data would make the file too large. Instead,
# these data are preprocessed, and a dF/F fluorescence signal extracted for each region-of-interest (ROI). To store the
# chain of computations necessary to describe this data processing pipeline, pynwb provides a "processing module" with
# interfaces that simplify and standardize the process of adding the steps in this provenance chain to the file:
ophys_module = nwbfile.create_processing_module(
    name="ophys_module",
    description="Processing module for 2P calcium responses",
)

########################################
# 6) First, we add an image segmentation interface to the module.  This interface implements a pre-defined schema and
# API that facilitates writing segmentation masks for ROI's:

image_segmentation_interface = ImageSegmentation(name="image_segmentation")

ophys_module.add(image_segmentation_interface)

plane_segmentation = image_segmentation_interface.create_plane_segmentation(
    name="plane_segmentation",
    description="Segmentation for imaging plane",
    imaging_plane=imaging_plane,
)

for cell_specimen_id in cell_specimen_ids:
    curr_name = cell_specimen_id
    curr_image_mask = dataset.get_roi_mask_array([cell_specimen_id])[0]
    plane_segmentation.add_roi(id=curr_name, image_mask=curr_image_mask)

########################################
# 7) Next, we add a dF/F  interface to the module.  This allows us to write the dF/F timeseries data associated with
# each ROI.

dff_interface = DfOverF(name="dff_interface")
ophys_module.add(dff_interface)

rt_region = plane_segmentation.create_roi_table_region(
    description="segmented cells with cell_specimen_ids",
)

dFF_series = dff_interface.create_roi_response_series(
    name="df_over_f",
    data=dFF,
    unit="NA",
    rois=rt_region,
    timestamps=timestamps,
)

########################################
# Now that we have created the data set, we can write the file to disk:
with NWBHDF5IO(save_file_name, mode="w") as io:
    io.write(nwbfile)

########################################
# For good measure, lets read the data back in and see if everything went as planned:
with NWBHDF5IO(save_file_name, mode="r") as io:
    nwbfile_in = io.read()



# -*- coding: utf-8 -*-
#
# sample documentation build configuration file, created by
# sphinx-quickstart on Mon Apr 16 21:22:43 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Not that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

import sphinx_rtd_theme


# -- Support building doc without install --------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# Get the project root dir, which is the parent parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(os.path.dirname(cwd))

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, os.path.join(project_root, 'src'))

import pynwb

# -- Autodoc configuration -----------------------------------------------------

autoclass_content = 'both'
autodoc_docstring_signature = True
autodoc_member_order = 'bysource'

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    'sphinx_gallery.gen_gallery',
    'sphinx_copybutton',
    "sphinxcontrib.jquery",  # can be removed as soon as the theme no longer depends on jQuery
]

from sphinx_gallery.sorting import ExplicitOrder
from sphinx_gallery.sorting import ExampleTitleSortKey


class CustomSphinxGallerySectionSortKey(ExampleTitleSortKey):
    """
    Define the key to be used to sort sphinx galleries sections

    :param src_dir : The source directory.
    :type srd_dir: str
    """
    # Define a partial ordered list of galleries for all subsections. Galleries not
    # listed here will be added in alphabetical order based on title after the
    # explicitly listed galleries
    GALLERY_ORDER = {
        'general': ['plot_file.py'],
        # Sort domain-specific tutorials based on domain to group tutorials belonging to the same domain
        'domain': [
            "ecephys.py",
            "ophys.py",
            "plot_icephys.py",
            "plot_icephys_pandas.py",
            "icephys.py",
            "plot_behavior.py",
            "images.py",
            "brain_observatory.py"
        ],
        'advanced_io': []
    }

    def __call__(self, filename):
        """
        Compute index to use for sorting galleries.

        Return the explicit index of the gallery if defined as part of self.GALLERY_ORDER
        and otherwise compute a score based on the title of the gallery to ensure galleries
        are sorted alphabetically by default
        """
        import string
        import math

        # Get the ordered list of gallery files for the current source dir
        explicit_order = self.GALLERY_ORDER.get(os.path.basename(self.src_dir), [])
        # If the file is in the explicit order then return its index
        if filename in explicit_order:
            sort_index = explicit_order.index(filename)
        # Else sort alphabetically based on the title by computing a corresponding
        # floating point index based on the characters of the titles
        else:
            title = super().__call__(filename)
            # map the characters of the title to a floating point number
            # based on the numerical index of the individual lowercase characters
            sort_index = len(explicit_order)  # all explicitly ordered galleries come first
            for i, v in enumerate(title.lower()):
                # get the index of the current character
                curr_index = (string.ascii_lowercase.index(v)
                              if v in string.ascii_lowercase
                              else len(string.ascii_lowercase))
                # shift the value based on its position in the title string and
                # add it to the sort_index value
                sort_index += curr_index / math.pow(10, ((i+1) * 2))
        return sort_index


sphinx_gallery_conf = {
    # path to your examples scripts
    'examples_dirs': ['../gallery'],
    # path where to save gallery generated examples
    'gallery_dirs': ['tutorials'],
    'subsection_order': ExplicitOrder(['../gallery/general', '../gallery/domain', '../gallery/advanced_io']),
    'backreferences_dir': 'gen_modules/backreferences',
    'min_reported_time': 5,
    'remove_config_comments': True,
    'within_subsection_order': CustomSphinxGallerySectionSortKey,
    'nested_sections': False,  # See issue https://github.com/sphinx-gallery/sphinx-gallery/issues/1152
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.11', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'h5py': ('https://docs.h5py.org/en/latest/', None),
    'hdmf': ('https://hdmf.readthedocs.io/en/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'dandi': ('https://dandi.readthedocs.io/en/stable/', None),
    'fsspec': ("https://filesystem-spec.readthedocs.io/en/latest/", None),
    'nwbwidgets': ("https://nwb-widgets.readthedocs.io/en/latest/", None),
    'nwb-overview': ("https://nwb-overview.readthedocs.io/en/latest/", None),
    'zarr': ("https://zarr.readthedocs.io/en/stable/", None),
    'hdmf-zarr': ("https://hdmf-zarr.readthedocs.io/en/latest/", None),
    'numcodecs': ("https://numcodecs.readthedocs.io/en/latest/", None),
}

extlinks = {
    'incf_lesson': ('https://training.incf.org/lesson/%s', '%s'),
    'incf_collection': ('https://training.incf.org/collection/%s', '%s'),
    'nwb_extension': ('https://github.com/nwb-extensions/%s', '%s'),
    'pynwb': ('https://github.com/NeurodataWithoutBorders/pynwb/%s', '%s'),
    'nwb_overview': ('https://nwb-overview.readthedocs.io/en/latest/%s', '%s'),
    'hdmf-docs': ('https://hdmf.readthedocs.io/en/stable/%s', '%s'),
    'dandi': ('https://www.dandiarchive.org/%s', '%s'),
    "nwbinspector": ("https://nwbinspector.readthedocs.io/en/dev/%s", "%s"),
    'hdmf-zarr': ('https://hdmf-zarr.readthedocs.io/en/latest/%s', '%s'),
}

nitpicky = True
nitpick_ignore = [('py:class', 'Intracomm'),
                  ('py:class', 'BaseStorageSpec')]

suppress_warnings = ["config.cache"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'source/index'
master_doc = 'index'

# General information about the project.
project = u'PyNWB'
copyright = u'2017-2024, Neurodata Without Borders'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = pynwb.__version__
# The full version, including alpha/beta/rc tags.
release = pynwb.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'test.py']

# List of patterns, relative to source directory, of modules to be
# excluded by apidoc when generating rst files.
apidoc_exclude = [
    "../../src/pynwb/retinotopy.py",
]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'default'
# html_theme = "sphinxdoc"
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # "style_nav_header_background": "#AFD2E8"
    "style_nav_header_background": "#000000"
}

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'css/custom.css',
]

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None
html_logo = 'figures/logo_pynwb_with_margin.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'figures/favicon_96.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'sampledoc'


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'print()reamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
# latex_documents = [
#  ('index', 'sample.tex', u'sample Documentation',
#   u'Kenneth Reitz', 'manual'),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = 'figures/logo_pynwb_with_margin.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#    ('index', 'sample', u'sample Documentation',
#     [u'Kenneth Reitz'], 1)
# ]
#
# If true, show URL addresses after external links.
# man_show_urls = False
#
#
# # -- Options for Texinfo output ------------------------------------------------
#
# # Grouping the document tree into Texinfo files. List of tuples
# # (source start file, target name, title, author,
# #  dir menu entry, description, category)
# texinfo_documents = [
#  ('index', 'sample', u'sample Documentation',
#   u'Kenneth Reitz', 'sample', 'One line description of project.',
#   'Miscellaneous'),
# ]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


# -- PyNWB sphinx extension ----------------------------------------------------

#
# see http://www.sphinx-doc.org/en/master/extdev/appapi.html
#

def run_apidoc(_):
    from sphinx.ext.apidoc import main as apidoc_main
    import os
    import sys
    out_dir = os.path.dirname(__file__)
    src_dir = os.path.join(out_dir, '../../src')
    sys.path.append(src_dir)
    apidoc_exclude_abs = [os.path.join(out_dir, f) for f in apidoc_exclude]
    apidoc_main(['-f', '-e', '--no-toc', '-o', out_dir, src_dir, *apidoc_exclude_abs])


from abc import abstractproperty

def skip(app, what, name, obj, skip, options):
    if isinstance(obj, abstractproperty) or getattr(obj, '__isabstractmethod__', False):
        return False
    elif name == "__getitem__":
        return False
    return skip


def setup(app):
    app.connect('builder-inited', run_apidoc)
    app.add_css_file("theme_overrides.css")  # overrides for wide tables in RTD theme
    app.connect("autodoc-skip-member", skip)



import numpy as np

from pynwb import TimeSeries
from pynwb.misc import IntervalSeries
from pynwb.behavior import (SpatialSeries, BehavioralEpochs, BehavioralEvents, BehavioralTimeSeries, PupilTracking,
                            EyeTracking, CompassDirection, Position)
from pynwb.testing import TestCase


class SpatialSeriesConstructor(TestCase):
    def test_init(self):
        sS = SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            bounds=[(-1,1),(-1,1),(-1,1)],
            reference_frame='reference_frame',
            timestamps=[1., 2., 3.]
        )
        self.assertEqual(sS.name, 'test_sS')
        self.assertEqual(sS.unit, 'meters')
        self.assertEqual(sS.bounds, [(-1,1),(-1,1),(-1,1)])
        self.assertEqual(sS.reference_frame, 'reference_frame')

    def test_set_unit(self):
        sS = SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            reference_frame='reference_frame',
            unit='degrees',
            timestamps=[1., 2., 3.]
        )
        self.assertEqual(sS.unit, 'degrees')

    def test_gt_3_cols(self):
        msg = ("SpatialSeries 'test_sS' has data shape (5, 4) which is not compliant with NWB 2.5 and greater. "
               "The second dimension should have length <= 3 to represent at most x, y, z.")
        with self.assertWarnsWith(UserWarning, msg):
            SpatialSeries(
                name="test_sS",
                data=np.ones((5, 4)),
                reference_frame="reference_frame",
                rate=30.
            )


class BehavioralEpochsConstructor(TestCase):
    def test_init(self):
        data = [0, 1, 0]
        iS = IntervalSeries(name='test_iS', data=data, timestamps=[1., 2., 3.])

        bE = BehavioralEpochs(iS)
        self.assertEqual(bE.interval_series['test_iS'], iS)


class BehavioralEventsConstructor(TestCase):
    def test_init(self):
        ts = TimeSeries(name='test_ts', data=np.ones((3, 2)), unit='unit', timestamps=[1., 2., 3.])

        bE = BehavioralEvents(ts)
        self.assertEqual(bE.time_series['test_ts'], ts)


class BehavioralTimeSeriesConstructor(TestCase):
    def test_init(self):
        ts = TimeSeries(name='test_ts', data=np.ones((3, 2)), unit='unit', timestamps=[1., 2., 3.])

        bts = BehavioralTimeSeries(ts)
        self.assertEqual(bts.time_series['test_ts'], ts)


class PupilTrackingConstructor(TestCase):
    def test_init(self):
        ts = TimeSeries(name='test_ts', data=np.ones((3, 2)), unit='unit', timestamps=[1., 2., 3.])

        pt = PupilTracking(ts)
        self.assertEqual(pt.time_series['test_ts'], ts)


class EyeTrackingConstructor(TestCase):
    def test_init(self):
        sS = SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            reference_frame='reference_frame',
            timestamps=[1., 2., 3.]
        )

        et = EyeTracking(sS)
        self.assertEqual(et.spatial_series['test_sS'], sS)


class CompassDirectionConstructor(TestCase):
    def test_init(self):
        sS = SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            reference_frame='reference_frame',
            timestamps=[1., 2., 3.]
        )

        cd = CompassDirection(sS)
        self.assertEqual(cd.spatial_series['test_sS'], sS)


class PositionConstructor(TestCase):
    def test_init(self):
        sS = SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            reference_frame='reference_frame',
            timestamps=[1., 2., 3.]
        )

        pc = Position(sS)
        self.assertEqual(pc.spatial_series.get('test_sS'), sS)



"""Utilities for creating a custom TypeMap for testing so that we don't use the global type map."""
import tempfile
from pynwb import get_type_map
from pynwb.spec import NWBNamespaceBuilder, export_spec


NAMESPACE_NAME = "test_core"


def create_test_extension(specs, container_classes, mappers=None):
    ns_builder = NWBNamespaceBuilder(
        name=NAMESPACE_NAME,
        version="0.1.0",
        doc="test extension",
    )
    ns_builder.include_namespace("core")

    output_dir = tempfile.TemporaryDirectory()
    export_spec(ns_builder, specs, output_dir.name)

    # this will copy the global pynwb TypeMap and add the extension types to the copy
    type_map = get_type_map(f"{output_dir.name}/{NAMESPACE_NAME}.namespace.yaml")
    for type_name, container_cls in container_classes.items():
        type_map.register_container_type(NAMESPACE_NAME, type_name, container_cls)
    if mappers:
        for type_name, mapper_cls in mappers.items():
            container_cls = container_classes[type_name]
            type_map.register_map(container_cls, mapper_cls)

    output_dir.cleanup()
    return type_map



import warnings

from pynwb.resources import HERD
from pynwb.testing import TestCase


class TestNWBContainer(TestCase):
    def test_constructor(self):
        """
        Test constructor
        """
        with warnings.catch_warnings(record=True):
            warnings.filterwarnings(
                "ignore",
                message=r"HERD is experimental .*",
                category=UserWarning,
            )
            er = HERD()
            self.assertIsInstance(er, HERD)



raise ImportError(
    "The pynwb.retinotopy module is deprecated. If you are interested in using these neurodata types, "
    "please create an issue on https://github.com/NeurodataWithoutBorders/nwb-schema/issues."
)

# import warnings
# from collections.abc import Iterable

# from hdmf.utils import docval, popargs, get_docval

# from . import register_class, CORE_NAMESPACE
# from .core import NWBDataInterface, NWBData


# class RetinotopyImage(NWBData):
#     """Gray-scale anatomical image of cortical surface. Array structure: [rows][columns]
#     """

#     __nwbfields__ = ('bits_per_pixel',
#                      'dimension',
#                      'format',
#                      'field_of_view')

#     @docval({'name': 'name', 'type': str, 'doc': 'Name of this retinotopy image'},
#             {'name': 'data', 'type': Iterable, 'doc': 'Data field.'},
#             {'name': 'bits_per_pixel', 'type': int,
#              'doc': 'Number of bits used to represent each value. This is necessary to determine maximum '
#                     '(white) pixel value.'},
#             {'name': 'dimension', 'type': Iterable, 'shape': (2, ),
#              'doc': 'Number of rows and columns in the image.'},
#             {'name': 'format', 'type': Iterable, 'doc': 'Format of image. Right now only "raw" supported.'},
#             {'name': 'field_of_view', 'type': Iterable, 'shape': (2, ), 'doc': 'Size of viewing area, in meters.'})
#     def __init__(self, **kwargs):
#         bits_per_pixel, dimension, format, field_of_view = popargs(
#             'bits_per_pixel', 'dimension', 'format', 'field_of_view', kwargs)
#         super().__init__(**kwargs)
#         self.bits_per_pixel = bits_per_pixel
#         self.dimension = dimension
#         self.format = format
#         self.field_of_view = field_of_view


# class FocalDepthImage(RetinotopyImage):
#     """Gray-scale image taken with same settings/parameters (e.g., focal depth,
#     wavelength) as data collection. Array format: [rows][columns].
#     """

#     __nwbfields__ = ('focal_depth', )

#     @docval(*get_docval(RetinotopyImage.__init__),
#             {'name': 'focal_depth', 'type': float, 'doc': 'Focal depth offset, in meters.'})
#     def __init__(self, **kwargs):
#         focal_depth = popargs('focal_depth', kwargs)
#         super().__init__(**kwargs)
#         self.focal_depth = focal_depth


# class RetinotopyMap(NWBData):
#     """Abstract two-dimensional map of responses to stimuli along a single response axis (e.g., altitude)
#     """

#     __nwbfields__ = ('field_of_view',
#                      'dimension')

#     @docval({'name': 'name', 'type': str, 'doc': 'the name of this axis map'},
#             {'name': 'data', 'type': Iterable, 'shape': (None, None), 'doc': 'data field.'},
#             {'name': 'field_of_view', 'type': Iterable, 'shape': (2, ), 'doc': 'Size of viewing area, in meters.'},
#             {'name': 'dimension', 'type': Iterable, 'shape': (2, ),
#              'doc': 'Number of rows and columns in the image'})
#     def __init__(self, **kwargs):
#         field_of_view, dimension = popargs('field_of_view', 'dimension', kwargs)
#         super().__init__(**kwargs)
#         self.field_of_view = field_of_view
#         self.dimension = dimension


# class AxisMap(RetinotopyMap):
#     """Abstract two-dimensional map of responses to stimuli along a single response axis (e.g., altitude) with unit
#     """

#     __nwbfields__ = ('unit', )

#     @docval(*get_docval(RetinotopyMap.__init__, 'name', 'data', 'field_of_view'),
#             {'name': 'unit', 'type': str, 'doc': 'Unit that axis data is stored in (e.g., degrees)'},
#             *get_docval(RetinotopyMap.__init__, 'dimension'))
#     def __init__(self, **kwargs):
#         unit = popargs('unit', kwargs)
#         super().__init__(**kwargs)
#         self.unit = unit


# @register_class('ImagingRetinotopy', CORE_NAMESPACE)
# class ImagingRetinotopy(NWBDataInterface):
#     """
#     Intrinsic signal optical imaging or widefield imaging for measuring retinotopy. Stores orthogonal
#     maps (e.g., altitude/azimuth; radius/theta) of responses to specific stimuli and a combined
#     polarity map from which to identify visual areas.
#     This group does not store the raw responses imaged during retinotopic mapping or the
#     stimuli presented, but rather the resulting phase and power maps after applying a Fourier
#     transform on the averaged responses.
#     Note: for data consistency, all images and arrays are stored in the format [row][column] and
#     [row, col], which equates to [y][x]. Field of view and dimension arrays may appear backward
#     (i.e., y before x).
#     """

#     __nwbfields__ = ({'name': 'sign_map', 'child': True},
#                      {'name': 'axis_1_phase_map', 'child': True},
#                      {'name': 'axis_1_power_map', 'child': True},
#                      {'name': 'axis_2_phase_map', 'child': True},
#                      {'name': 'axis_2_power_map', 'child': True},
#                      {'name': 'focal_depth_image', 'child': True},
#                      {'name': 'vasculature_image', 'child': True},
#                      'axis_descriptions')

#     @docval({'name': 'sign_map', 'type': RetinotopyMap,
#              'doc': 'Sine of the angle between the direction of the gradient in axis_1 and axis_2.'},
#             {'name': 'axis_1_phase_map', 'type': AxisMap,
#              'doc': 'Phase response to stimulus on the first measured axis.'},
#             {'name': 'axis_1_power_map', 'type': AxisMap,
#              'doc': 'Power response on the first measured axis. Response is scaled so 0.0 is no power in '
#                     'the response and 1.0 is maximum relative power.'},
#             {'name': 'axis_2_phase_map', 'type': AxisMap,
#              'doc': 'Phase response to stimulus on the second measured axis.'},
#             {'name': 'axis_2_power_map', 'type': AxisMap,
#              'doc': 'Power response on the second measured axis. Response is scaled so 0.0 is no '
#                      'power in the response and 1.0 is maximum relative power.'},
#             {'name': 'axis_descriptions', 'type': Iterable, 'shape': (2, ),
#              'doc': 'Two-element array describing the contents of the two response axis fields. '
#                     'Description should be something like ["altitude", "azimuth"] or ["radius", "theta"].'},
#             {'name': 'focal_depth_image', 'type': FocalDepthImage,
#              'doc': 'Gray-scale image taken with same settings/parameters (e.g., focal depth, wavelength) '
#                     'as data collection. Array format - [rows][columns].'},
#             {'name': 'vasculature_image', 'type': RetinotopyImage,
#              'doc': 'Gray-scale anatomical image of cortical surface. Array structure - [rows][columns].'},
#             {'name': 'name', 'type': str, 'doc': 'the name of this container', 'default': 'ImagingRetinotopy'})
#     def __init__(self, **kwargs):
#         axis_1_phase_map, axis_1_power_map, axis_2_phase_map, axis_2_power_map, axis_descriptions, \
#             focal_depth_image, sign_map, vasculature_image = popargs(
#                 'axis_1_phase_map', 'axis_1_power_map', 'axis_2_phase_map', 'axis_2_power_map',
#                 'axis_descriptions', 'focal_depth_image', 'sign_map', 'vasculature_image', kwargs)
#         super().__init__(**kwargs)
#         warnings.warn("The ImagingRetinotopy class currently cannot be written to or read from a file. "
#                       "This is a known bug and will be fixed in a future release of PyNWB.")
#         self.axis_1_phase_map = axis_1_phase_map
#         self.axis_1_power_map = axis_1_power_map
#         self.axis_2_phase_map = axis_2_phase_map
#         self.axis_2_power_map = axis_2_power_map
#         self.axis_descriptions = axis_descriptions
#         self.focal_depth_image = focal_depth_image
#         self.sign_map = sign_map
#         self.vasculature_image = vasculature_image



from typing import Optional

import numpy as np

from pynwb.icephys import (
    IntracellularElectrode,
    VoltageClampSeries,
    VoltageClampStimulusSeries,
    IntracellularRecordingsTable,
    CurrentClampSeries,
    CurrentClampStimulusSeries,
    IZeroClampSeries,
)

from .utils import name_generator
from .device import mock_Device
from ... import NWBFile
from ...device import Device


def mock_IntracellularElectrode(
    name: Optional[str] = None,
        description: str = "description",
        device: Optional[Device] = None,
        nwbfile: Optional[NWBFile] = None,
) -> IntracellularElectrode:
    intracellular_electrode = IntracellularElectrode(
        name=name or name_generator("IntracellularElectrode"),
        description=description,
        device=device or mock_Device(nwbfile=nwbfile),
    )

    if nwbfile is not None:
        nwbfile.add_icephys_electrode(intracellular_electrode)

    return intracellular_electrode


def mock_VoltageClampStimulusSeries(
    name: Optional[str] = None,
    data=None,
    rate: float = 100_000.,
    electrode: Optional[IntracellularElectrode] = None,
    gain: float = 0.02,
    timestamps=None,
    starting_time: Optional[float] = None,
    nwbfile: Optional[NWBFile] = None,
) -> VoltageClampStimulusSeries:
    voltage_clamp_stimulus_series = VoltageClampStimulusSeries(
        name=name or name_generator("VoltageClampStimulusSeries"),
        data=data or np.ones((30,)),
        rate=None if timestamps else rate,
        electrode=electrode or mock_IntracellularElectrode(nwbfile=nwbfile),
        gain=gain,
        timestamps=timestamps,
        starting_time=starting_time,
    )

    if nwbfile is not None:
        nwbfile.add_stimulus(voltage_clamp_stimulus_series)

    return voltage_clamp_stimulus_series


def mock_VoltageClampSeries(
    name: Optional[str] = None,
    data=None,
    conversion: float = 1.0,
    resolution: float = np.nan,
    starting_time: Optional[float] = None,
    rate: Optional[float] = 100_000.0,
    electrode: Optional[IntracellularElectrode] = None,
    gain: float = 0.02,
    capacitance_slow: float = 100e-12,
    resistance_comp_correction: float = 70.0,
    nwbfile: Optional[NWBFile] = None,
) -> VoltageClampSeries:
    voltage_clamp_series = VoltageClampSeries(
        name=name if name is not None else name_generator("VoltageClampSeries"),
        data=data if data is not None else np.ones((30,)),
        conversion=conversion,
        resolution=resolution,
        starting_time=starting_time,
        rate=rate,
        electrode=electrode or mock_IntracellularElectrode(nwbfile=nwbfile),
        gain=gain,
        capacitance_slow=capacitance_slow,
        resistance_comp_correction=resistance_comp_correction,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(voltage_clamp_series)

    return voltage_clamp_series


def mock_CurrentClampSeries(
    name: Optional[str] = None,
    data=None,
    electrode: Optional[IntracellularElectrode] = None,
    gain: float = 0.02,
    stimulus_description: str = "N/A",
    bias_current=None,
    bridge_balance=None,
    capacitance_compensation=None,
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time: Optional[float] = None,
    rate: Optional[float] = 100_000.,
    comments: str = "no comments",
    description: str = "no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit: str = "volts",
    nwbfile: Optional[NWBFile] = None,
) -> CurrentClampSeries:
    current_clamp_series = CurrentClampSeries(
        name=name if name is not None else name_generator("CurrentClampSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(nwbfile=nwbfile),
        gain=gain,
        stimulus_description=stimulus_description,
        bias_current=bias_current,
        bridge_balance=bridge_balance,
        capacitance_compensation=capacitance_compensation,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(current_clamp_series)

    return current_clamp_series


def mock_CurrentClampStimulusSeries(
    name: Optional[str] = None,
    data=None,
    electrode=None,
    gain=0.02,
    stimulus_description="N/A",
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=100_000.,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit="amperes",
    nwbfile: Optional[NWBFile] = None,
) -> CurrentClampStimulusSeries:
    current_clamp_stimulus_series = CurrentClampStimulusSeries(
        name=name or name_generator("CurrentClampStimulusSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(nwbfile=nwbfile),
        gain=gain,
        stimulus_description=stimulus_description,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )

    if nwbfile is not None:
        nwbfile.add_stimulus(current_clamp_stimulus_series)

    return current_clamp_stimulus_series


def mock_IZeroClampSeries(
    name: Optional[str] = None,
    data=None,
    electrode: Optional[IntracellularElectrode] = None,
    gain=.02,
    stimulus_description="N/A",
    resolution=-1.0,
    conversion=1.0,
    timestamps=None,
    starting_time=None,
    rate=100_000.,
    comments="no comments",
    description="no description",
    control=None,
    control_description=None,
    sweep_number=None,
    offset=0.0,
    unit="volts",
    nwbfile: Optional[NWBFile] = None,
) -> IZeroClampSeries:
    izero_clamp_series = IZeroClampSeries(
        name=name or name_generator("IZeroClampSeries"),
        data=data if data is not None else np.ones((30,)),
        electrode=electrode or mock_IntracellularElectrode(nwbfile=nwbfile),
        gain=gain,
        stimulus_description=stimulus_description,
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
        sweep_number=sweep_number,
        offset=offset,
        unit=unit,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(izero_clamp_series)

    return izero_clamp_series


def mock_IntracellularRecordingsTable(
    n_rows: int = 5, nwbfile: Optional[NWBFile] = None
) -> IntracellularRecordingsTable:
    irt = IntracellularRecordingsTable()
    for _ in range(n_rows):
        electrode = mock_IntracellularElectrode(nwbfile=nwbfile)
        irt.add_recording(
            electrode=electrode,
            stimulus=mock_VoltageClampStimulusSeries(electrode=electrode, nwbfile=nwbfile),
            response=mock_VoltageClampSeries(electrode=electrode, nwbfile=nwbfile),
        )

    if nwbfile is not None:
        nwbfile.intracellular_recordings = irt

    return irt



# -*- coding: utf-8 -*-
"""
.. _icephys_tutorial_new:

Intracellular Electrophysiology
===============================

The following tutorial describes storage of intracellular electrophysiology data in NWB.
NWB supports storage of the time series describing the stimulus and response, information
about the electrode and device used, as well as metadata about the organization of the
experiment.

.. note:: For a video tutorial on intracellular electrophysiology in NWB see also the
       :incf_lesson:`Intracellular electrophysiology basics in NWB <intracellular-electrophysiology-basics-nwb>` and
       :incf_lesson:`Intracellular ephys metadata <intracellular-electrophysiology-structured-metadata-nwb>`
       tutorials as part of the :incf_collection:`NWB Course <neurodata-without-borders-neurophysiology-nwbn>`
       at the INCF Training Space.

.. figure:: ../../figures/plot_icephys_table_hierarchy.png
    :figwidth: 100%
    :alt: Intracellular electrophysiology metadata table hierarchy

    Illustration of the hierarchy of metadata tables used to describe the organization of
    intracellular electrophysiology experiments.

Recordings of intracellular electrophysiology stimuli and responses are represented
with subclasses of :py:class:`~pynwb.icephys.PatchClampSeries` using the
:py:class:`~pynwb.icephys.IntracellularElectrode` and :py:class:`~pynwb.device.Device`
type to describe the electrode and device used.

To describe the organization of intracellular experiments, the metadata is organized
hierarchically in a sequence of tables. All of the tables are so-called DynamicTables
enabling users to add columns for custom metadata.

- :py:class:`~pynwb.icephys.IntracellularRecordingsTable` relates electrode, stimulus
  and response pairs and describes metadata specific to individual recordings.
- :py:class:`~pynwb.icephys.SimultaneousRecordingsTable` groups intracellular
  recordings from the :py:class:`~pynwb.icephys.IntracellularRecordingsTable`
  together that were recorded simultaneously from different electrodes and/or cells
  and describes metadata that is constant across the simultaneous recordings.
  In practice a simultaneous recording is often  also referred to as a sweep.
- :py:class:`~pynwb.icephys.SequentialRecordingsTable` groups simultaneously
  recorded intracellular recordings from the
  :py:class:`~pynwb.icephys.SimultaneousRecordingsTable` together and describes
  metadata that is constant across the simultaneous recordings. In practice a
  sequential recording is often also referred to as a sweep sequence. A common
  use of sequential recordings is to group together simultaneous recordings
  where a sequence of stimuli of the same type with varying parameters
  have been presented in a sequence (e.g., a sequence of square waveforms with
  varying amplitude).
- :py:class:`~pynwb.icephys.RepetitionsTable` groups sequential recordings from
  the :py:class:`~pynwb.icephys.SequentialRecordingsTable`. In practice a
  repetition is often also referred to a run. A typical use of the
  :py:class:`~pynwb.icephys.RepetitionsTable` is to group sets of different stimuli
  that are applied in sequence that may be repeated.
- :py:class:`~pynwb.icephys.ExperimentalConditionsTable` groups repetitions of
  intracellular recording from the :py:class:`~pynwb.icephys.RepetitionsTable`
  together that belong to the same experimental conditions.

Storing data in hierarchical tables has the advantage that it allows us to avoid duplication of
metadata. E.g., for a single experiment we only need to describe the metadata that is constant
across an experimental condition as a single row in the :py:class:`~pynwb.icephys.ExperimentalConditionsTable`
without having to replicate the same information across all repetitions and sequential-, simultaneous-, and
individual intracellular recordings. For analysis, this means that we can easily focus on individual
aspects of an experiment while still being able to easily access information about information from
related tables.

.. note:: All of the above mentioned metadata tables are optional and are created automatically
          by the :py:class:`~pynwb.file.NWBFile` class the first time data is being
          added to a table via the corresponding add functions. However, as tables at higher
          levels of the hierarchy link to the other tables that are lower in the hierarchy,
          we may only exclude tables from the top of the hierarchy. This means, for example,
          a file containing a :py:class:`~pynwb.icephys.SimultaneousRecordingsTable` then
          must also always contain a corresponding :py:class:`~pynwb.icephys.IntracellularRecordingsTable`.
"""

#####################################################################
# Imports used in the tutorial
# ------------------------------

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_icephys.png'
# Standard Python imports
from datetime import datetime
from uuid import uuid4

import numpy as np
import pandas
from dateutil.tz import tzlocal

# Set pandas rendering option to avoid very wide tables in the html docs
pandas.set_option("display.max_colwidth", 30)
pandas.set_option("display.max_rows", 10)

# Import I/O class used for reading and writing NWB files
# Import main NWB file class
from pynwb import NWBHDF5IO, NWBFile

# Import additional core datatypes used in the example
from pynwb.core import DynamicTable, VectorData
from pynwb.base import TimeSeriesReference, TimeSeriesReferenceVectorData

# Import icephys TimeSeries types used
from pynwb.icephys import VoltageClampSeries, VoltageClampStimulusSeries

#####################################################################
# A brief example
# ---------------
#
# The following brief example provides a quick overview of the main steps
# to create an NWBFile for intracelluar electrophysiology data. We then
# discuss the individual steps in more detail afterwards.
#
# .. note:
#    To avoid collisions between this example script and the more
#    detailed discussion we prefix all variables in the example
#    script with ``ex_``.

# Create an ICEphysFile
ex_nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier=str(uuid4()),
    session_start_time=datetime.now(tzlocal()),
    experimenter="Baggins, Bilbo",
    lab="Bag End Laboratory",
    institution="University of Middle Earth at the Shire",
    experiment_description="I went on an adventure with thirteen dwarves "
    "to reclaim vast treasures.",
    session_id="LONELYMTN",
)

# Add a device
ex_device = ex_nwbfile.create_device(name="Heka ITC-1600")

# Add an intracellular electrode
ex_electrode = ex_nwbfile.create_icephys_electrode(
    name="elec0", description="a mock intracellular electrode", device=ex_device
)

# Create an ic-ephys stimulus
ex_stimulus = VoltageClampStimulusSeries(
    name="stimulus",
    data=[1, 2, 3, 4, 5],
    starting_time=123.6,
    rate=10e3,
    electrode=ex_electrode,
    gain=0.02,
)

# Create an ic-response
ex_response = VoltageClampSeries(
    name="response",
    data=[0.1, 0.2, 0.3, 0.4, 0.5],
    conversion=1e-12,
    resolution=np.nan,
    starting_time=123.6,
    rate=20e3,
    electrode=ex_electrode,
    gain=0.02,
    capacitance_slow=100e-12,
    resistance_comp_correction=70.0,
)

# (A) Add an intracellular recording to the file
#     NOTE: We can optionally define time-ranges for the stimulus/response via
#     the corresponding optional _start_index and _index_count parameters.
#     NOTE: It is allowed to add a recording with just a stimulus or a response
#     NOTE: We can  add custom columns to any of our tables in steps (A)-(E)
ex_ir_index = ex_nwbfile.add_intracellular_recording(
    electrode=ex_electrode, stimulus=ex_stimulus, response=ex_response
)

# (B) Add a list of sweeps to the simultaneous recordings table
ex_sweep_index = ex_nwbfile.add_icephys_simultaneous_recording(
    recordings=[
        ex_ir_index,
    ]
)

# (C) Add a list of simultaneous recordings table indices as a sequential recording
ex_sequence_index = ex_nwbfile.add_icephys_sequential_recording(
    simultaneous_recordings=[
        ex_sweep_index,
    ],
    stimulus_type="square",
)

# (D) Add a list of sequential recordings table indices as a repetition
run_index = ex_nwbfile.add_icephys_repetition(
    sequential_recordings=[
        ex_sequence_index,
    ]
)

# (E) Add a list of repetition table indices as a experimental condition
ex_nwbfile.add_icephys_experimental_condition(
    repetitions=[
        run_index,
    ]
)

# Write our test file
ex_testpath = "ex_test_icephys_file.nwb"
with NWBHDF5IO(ex_testpath, "w") as io:
    io.write(ex_nwbfile)

# Read the data back in
with NWBHDF5IO(ex_testpath, "r") as io:
    infile = io.read()

# Optionally plot the organization of our example NWB file
try:
    import matplotlib.pyplot as plt
    from hdmf_docutils.doctools.render import (
        HierarchyDescription,
        NXGraphHierarchyDescription,
    )

    ex_file_hierarchy = HierarchyDescription.from_hdf5(ex_testpath)
    ex_file_graph = NXGraphHierarchyDescription(ex_file_hierarchy)
    ex_fig = ex_file_graph.draw(
        show_plot=False,
        figsize=(12, 16),
        label_offset=(0.0, 0.0065),
        label_font_size=10,
    )
    plt.show()
except ImportError:  # ignore in case hdmf_docutils is not installed
    pass

#####################################################################
# Now that we have seen a brief example, we are going to start from the beginning and
# go through each of the steps in more detail in the following sections.


#####################################################################
# Creating an NWB file for Intracellular electrophysiology
# --------------------------------------------------------
#
# When creating an NWB file, the first step is to create the :py:class:`~pynwb.file.NWBFile`. The first
# argument is a brief description of the dataset.

# Create the file
nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier=str(uuid4()),
    session_start_time=datetime.now(tzlocal()),
    experimenter=[
        "Baggins, Bilbo",
    ],
    lab="Bag End Laboratory",
    institution="University of Middle Earth at the Shire",
    experiment_description="I went on an adventure to reclaim vast treasures.",
    session_id="LONELYMTN001",
)

#####################################################################
# Device metadata
# ^^^^^^^^^^^^^^^
#
# Device metadata is represented by :py:class:`~pynwb.device.Device` objects.
# To create a device, you can use the :py:class:`~pynwb.file.NWBFile` instance method
# :py:meth:`~pynwb.file.NWBFile.create_device`.

device = nwbfile.create_device(name="Heka ITC-1600")

#####################################################################
# Electrode metadata
# ^^^^^^^^^^^^^^^^^^
#
# Intracellular electrode metadata is represented by :py:class:`~pynwb.icephys.IntracellularElectrode` objects.
# To create an electrode group, you can use the :py:class:`~pynwb.file.NWBFile` instance method
# :py:meth:`~pynwb.file.NWBFile.create_icephys_electrode`.

electrode = nwbfile.create_icephys_electrode(
    name="elec0", description="a mock intracellular electrode", device=device
)

#####################################################################
# Stimulus and response data
# ^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Intracellular stimulus and response data are represented with subclasses of
# :py:class:`~pynwb.icephys.PatchClampSeries`. A stimulus is described by a
# time series representing voltage or current stimulation with a particular
# set of parameters. There are two classes for representing stimulus data:
#
# - :py:class:`~pynwb.icephys.VoltageClampStimulusSeries`
# - :py:class:`~pynwb.icephys.CurrentClampStimulusSeries`
#
# The response is then described by a time series representing voltage or current
# recorded from a single cell using a single intracellular electrode via one of
# the following classes:
#
# - :py:class:`~pynwb.icephys.VoltageClampSeries`
# - :py:class:`~pynwb.icephys.CurrentClampSeries`
# - :py:class:`~pynwb.icephys.IZeroClampSeries`
#
# Below we create a simple example stimulus/response recording data pair.

# Create an example icephys stimulus.
stimulus = VoltageClampStimulusSeries(
    name="ccss",
    data=[1, 2, 3, 4, 5],
    starting_time=123.6,
    rate=10e3,
    electrode=electrode,
    gain=0.02,
    sweep_number=np.uint64(15),
)

# Create and icephys response
response = VoltageClampSeries(
    name="vcs",
    data=[0.1, 0.2, 0.3, 0.4, 0.5],
    conversion=1e-12,
    resolution=np.nan,
    starting_time=123.6,
    rate=20e3,
    electrode=electrode,
    gain=0.02,
    capacitance_slow=100e-12,
    resistance_comp_correction=70.0,
    sweep_number=np.uint64(15),
)

#####################################################################
# Adding an intracellular recording
# ---------------------------------
#
# As mentioned earlier, intracellular recordings are organized in the
# :py:class:`~pynwb.icephys.IntracellularRecordingsTable` which relates electrode, stimulus
# and response pairs and describes metadata specific to individual recordings.
#
# .. figure:: ../../figures/plot_icephys_intracellular_recordings_table.png
#    :figwidth: 90%
#    :alt: IntracellularRecordingsTable
#
#    Illustration of the structure of the IntracellularRecordingsTable
#
# We can add an intracellular recording to the file via :py:meth:`~pynwb.file.NWBFile.add_intracellular_recording`.
# The function will record the data in the :py:class:`~pynwb.icephys.IntracellularRecordingsTable`
# and add the given electrode, stimulus, or response to the NWBFile object if necessary.
# Any time we add a row to one of our tables, the corresponding add function (here
# :py:meth:`~pynwb.file.NWBFile.add_intracellular_recording`) returns the integer index of the
# newly created row. The ``rowindex`` is used in subsequent tables that reference rows in our table.

rowindex = nwbfile.add_intracellular_recording(
    electrode=electrode, stimulus=stimulus, response=response, id=10
)

#####################################################################
# .. note:: Since :py:meth:`~pynwb.file.NWBFile.add_intracellular_recording` can automatically add
#          the objects to the NWBFile we do not need to separately call
#          :py:meth:`~pynwb.file.NWBFile.add_stimulus` and :py:meth:`~pynwb.file.NWBFile.add_acquisition`
#          to add our stimulus and response, but it is still fine to do so.
#
# .. note:: The ``id`` parameter in the call is optional and if the ``id`` is omitted then PyNWB will
#          automatically number recordings in sequences (i.e., id is the same as the rowindex)
#
# .. note:: The IntracellularRecordigns, SimultaneousRecordings, SequentialRecordingsTable,
#          RepetitionsTable and ExperimentalConditionsTable tables all enforce unique ids
#          when adding rows. I.e., adding an intracellular recording with the same id twice
#          results in a ValueError.
#

#####################################################################
# .. note:: We may optionally also specify the relevant time range for a stimulus and/or response as part of
#           the intracellular_recording. This is useful, e.g., in case where the recording of the stimulus
#           and response do not align (e.g., in case the recording of the response started before
#           the recording of the stimulus).

rowindex2 = nwbfile.add_intracellular_recording(
    electrode=electrode,
    stimulus=stimulus,
    stimulus_start_index=1,
    stimulus_index_count=3,
    response=response,
    response_start_index=2,
    response_index_count=3,
    id=11,
)

#####################################################################
# .. note:: A recording may optionally also consist of just an electrode and stimulus or electrode
#           and response, but at least one of stimulus or response is required. If either stimulus or response
#           is missing, then the stimulus and response are internally set to the same TimeSeries and the
#           start_index and index_count for the missing parameter are set to -1. When retrieving
#           data from the :py:class:`~pynwb.base.TimeSeriesReferenceVectorData`, the missing values
#           will be represented via masked numpy arrays, i.e., as masked values in a
#           ``numpy.ma.masked_array`` or as a ``np.ma.core.MaskedConstant``.

rowindex3 = nwbfile.add_intracellular_recording(
    electrode=electrode, response=response, id=12
)

#####################################################################
# .. warning:: For brevity we reused in the above example the same response and stimulus in
#        all rows of the intracellular_recordings. While this is allowed, in most practical
#        cases the stimulus and response will change between intracellular_recordings.

#####################################################################
# Adding custom columns to the intracellular recordings table
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# We can add a column to the main intracellular recordings table as follows.

nwbfile.intracellular_recordings.add_column(
    name="recording_tag",
    data=["A1", "A2", "A3"],
    description="String with a recording tag",
)

#####################################################################
# The :py:class:`~pynwb.icephys.IntracellularRecordingsTable` table is not just a ``DynamicTable``
# but an ``AlignedDynamicTable``. The ``AlignedDynamicTable`` type is itself a ``DynamicTable``
# that may contain an arbitrary number of additional ``DynamicTable``, each of which defines
# a "category". This is similar to a table with "sub-headings". In the case of the
# :py:class:`~pynwb.icephys.IntracellularRecordingsTable`, we have three predefined categories,
# i.e., electrodes, stimuli, and responses. We can also dynamically add new categories to
# the table. As each category corresponds to a ``DynamicTable``, this means we have to create a
# new ``DynamicTable`` and add it to our table.

# Create a new DynamicTable for our category that contains a location column of type VectorData
location_column = VectorData(
    name="location",
    data=["Mordor", "Gondor", "Rohan"],
    description="Recording location in Middle Earth",
)

lab_category = DynamicTable(
    name="recording_lab_data",
    description="category table for lab-specific recording metadata",
    colnames=[
        "location",
    ],
    columns=[
        location_column,
    ],
)
# Add the table as a new category to our intracellular_recordings
nwbfile.intracellular_recordings.add_category(category=lab_category)
# Note, the name of the category is name of the table, i.e., 'recording_lab_data'

#####################################################################
# .. note:: In an ``AlignedDynamicTable`` all category tables MUST align with the main table,
#           i.e., all tables must have the same number of rows and rows are expected to
#           correspond to each other by index

#####################################################################
# We can also add custom columns to any of the subcategory tables, i.e.,
# the electrodes, stimuli, and responses tables, and any custom subcategory tables.
# All we need to do is indicate the name of the category we want to add the column to.

nwbfile.intracellular_recordings.add_column(
    name="voltage_threshold",
    data=[0.1, 0.12, 0.13],
    description="Just an example column on the electrodes category table",
    category="electrodes",
)

#####################################################################
# Adding stimulus templates
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# One predefined subcategory column is the ``stimulus_template`` column in the stimuli table. This column is
# used to store template waveforms of stimuli in addition to the actual recorded stimulus that is stored in the
# ``stimulus`` column. The ``stimulus_template`` column contains an idealized version of the template waveform used as
# the stimulus. This can be useful as a noiseless version of the stimulus for data analysis or to validate that the
# recorded stimulus matches the expected waveform of the template. Similar to the ``stimulus`` and ``response``
# columns, we can specify a relevant time range.

stimulus_template = VoltageClampStimulusSeries(
    name="ccst",
    data=[0, 1, 2, 3, 4],
    starting_time=0.0,
    rate=10e3,
    electrode=electrode,
    gain=0.02,
)
nwbfile.add_stimulus_template(stimulus_template)

nwbfile.intracellular_recordings.add_column(
    name="stimulus_template",
    data=[TimeSeriesReference(0, 5, stimulus_template),  # (start_index, index_count, stimulus_template)
          TimeSeriesReference(1, 3, stimulus_template),
          TimeSeriesReference.empty(stimulus_template)],  # if there was no data for that recording, use empty reference
    description="Column storing the reference to the stimulus template for the recording (rows).",
    category="stimuli",
    col_cls=TimeSeriesReferenceVectorData
)

# we can also add stimulus template data as follows
rowindex = nwbfile.add_intracellular_recording(
    electrode=electrode,
    stimulus=stimulus,
    stimulus_template=stimulus_template,  # the full time range of the stimulus template will be used unless specified
    recording_tag='A4',
    recording_lab_data={'location': 'Isengard'},
    electrode_metadata={'voltage_threshold': 0.14},
    id=13,
)

#####################################################################
# .. note:: If a stimulus template column exists but there is no stimulus template data for that recording, then
#           :py:meth:`~pynwb.file.NWBFile.add_intracellular_recording` will internally set the stimulus template to the
#           provided stimulus or response TimeSeries and the start_index and index_count for the missing parameter are
#           set to -1. The missing values will be represented via masked numpy arrays.

#####################################################################
# .. note:: Since stimulus templates are often reused across many recordings, the timestamps in the templates are not
#           usually aligned with the recording nor with the reference time of the file. The timestamps often start
#           at 0 and are relative to the time of the application of the stimulus.

#####################################################################
# Add a simultaneous recording
# ---------------------------------
#
# Before adding a simultaneous recording, we will take a brief discourse to illustrate
# how we can add custom columns to tables before and after we have populated the table with data
#
# Define a custom column for a simultaneous recording before populating the table
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Before we add a simultaneous recording, let's create a custom data column in our
# :py:class:`~pynwb.icephys.SimultaneousRecordingsTable`. We can create columns at the
# beginning (i.e., before we populate the table with rows/data) or we can add columns
# after we have already populated the table with rows. Here we will show the former.
# For this, we first need to get access to our table.

print(nwbfile.icephys_simultaneous_recordings)

#####################################################################
# The :py:class:`~pynwb.icephys.SimultaneousRecordingsTable` is optional, and since we have
# not populated it with any data yet, we can see that the table does not actually exist yet.
# In order to make sure the table is being created we can use
# :py:meth:`~pynwb.file.NWBFile.get_icephys_simultaneous_recordings`, which ensures
# that the table is being created if it does not exist yet.

icephys_simultaneous_recordings = nwbfile.get_icephys_simultaneous_recordings()
icephys_simultaneous_recordings.add_column(
    name="simultaneous_recording_tag",
    description="A custom tag for simultaneous_recordings",
)
print(icephys_simultaneous_recordings.colnames)

#####################################################################
# As we can see, we now have successfully created a new custom column.
#
# .. note:: The same process applies to all our other tables as well. We can use the
#         corresponding :py:meth:`~pynwb.file.NWBFile.get_intracellular_recordings`,
#         :py:meth:`~pynwb.file.NWBFile.get_icephys_sequential_recordings`,
#         :py:meth:`~pynwb.file.NWBFile.get_icephys_repetitions` functions instead.
#         In general, we can always use the get functions instead of accessing the property
#         of the file.
#

#####################################################################
# Add a simultaneous recording
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Add a single simultaneous recording consisting of a set of intracellular recordings.
# Again, setting the id for a simultaneous recording is optional. The recordings
# argument of the :py:meth:`~pynwb.file.NWBFile.add_icephys_simultaneous_recording` function
# here is simply a list of ints with the indices of the corresponding rows in
# the :py:class:`~pynwb.icephys.IntracellularRecordingsTable`
#
# .. note:: Since we created our custom ``simultaneous_recording_tag column`` earlier,
#           we now also need to populate this custom field for every row we add to
#           the :py:class:`~pynwb.icephys.SimultaneousRecordingsTable`.
#

rowindex = nwbfile.add_icephys_simultaneous_recording(
    recordings=[rowindex, rowindex2, rowindex3],
    id=12,
    simultaneous_recording_tag="LabTag1",
)

#####################################################################
# .. note:: The ``recordings`` argument is the list of indices of the rows in the
#           :py:class:`~pynwb.icephys.IntracellularRecordingsTable` that we want
#           to reference. The indices are determined by the order in which we
#           added the elements to the table. If we don't know the row indices,
#           but only the ids of the relevant intracellular recordings, then
#           we can search for them as follows:

temp_row_indices = nwbfile.intracellular_recordings.id == [10, 11]
print(temp_row_indices)

#####################################################################
# .. note::  The same is true for our other tables as well, i.e., referencing is
#            done always by indices of rows (NOT ids). If we only know ids then we
#            can search for them in the same manner on the other tables as well,
#            e.g,. via  ``nwbfile.simultaneous_recordings.id == 15``. In the search
#            we can use a list of integer ids or a single int.

#####################################################################
# Define a custom column for a simultaneous recording after adding rows
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Depending on the lab workflow, it may be useful to add complete columns to a table
# after we have already populated the table with rows. We can do this the same way as
# before, only now we need to provide a data array to populate the values for
# the existing rows. E.g.:

nwbfile.icephys_simultaneous_recordings.add_column(
    name="simultaneous_recording_type",
    description="Description of the type of simultaneous_recording",
    data=[
        "SimultaneousRecordingType1",
    ],
)

#####################################################################
# Add a sequential recording
# --------------------------
#
# Add a single sequential recording consisting of a set of simultaneous recordings.
# Again, setting the id for a sequential recording is optional. Also this table is
# optional and will be created automatically by NWBFile. The ``simultaneous_recordings``
# argument of the :py:meth:`~pynwb.file.NWBFile.add_icephys_sequential_recording` function
# here is simply a list of ints with the indices of the corresponding rows in
# the :py:class:`~pynwb.icephys.SimultaneousRecordingsTable`.

rowindex = nwbfile.add_icephys_sequential_recording(
    simultaneous_recordings=[0], stimulus_type="square", id=15
)

#####################################################################
# Add a repetition
# ----------------
#
# Add a single repetition consisting of a set of sequential recordings. Again, setting
# the id for a repetition is optional. Also this table is optional and will be created
# automatically by NWBFile. The ``sequential_recordings argument`` of the
# :py:meth:`~pynwb.file.NWBFile.add_icephys_repetition` function here is simply
# a list of ints with the indices of the corresponding rows in
# the :py:class:`~pynwb.icephys.SequentialRecordingsTable`.

rowindex = nwbfile.add_icephys_repetition(sequential_recordings=[0], id=17)

#####################################################################
# Add an experimental condition
# -----------------------------
#
# Add a single experimental condition consisting of a set of repetitions. Again,
# setting the id for a condition is optional. Also this table is optional and
# will be created automatically by NWBFile. The ``repetitions`` argument of
# the :py:meth:`~pynwb.file.NWBFile.add_icephys_experimental_condition` function again is
# simply a list of ints with the indices of the correspondingto rows in the
# :py:class:`~pynwb.icephys.RepetitionsTable`.

rowindex = nwbfile.add_icephys_experimental_condition(repetitions=[0], id=19)

#####################################################################
# As mentioned earlier, to add additional columns to any of the tables, we can
# use the ``.add_column`` function on the corresponding table after they have
# been created.

nwbfile.icephys_experimental_conditions.add_column(
    name="tag",
    data=np.arange(1),
    description="integer tag for a experimental condition",
)

#####################################################################
# When we add new items, then we now also need to set the values for the new column, e.g.:

rowindex = nwbfile.add_icephys_experimental_condition(repetitions=[0], id=21, tag=3)

#####################################################################
# Read/write the NWBFile
# -----------------------------
#

# Write our test file
testpath = "test_icephys_file.nwb"
with NWBHDF5IO(testpath, "w") as io:
    io.write(nwbfile)

# Read the data back in
with NWBHDF5IO(testpath, "r") as io:
    infile = io.read()


#####################################################################
# Accessing the tables
# -----------------------------
#
# All of the icephys metadata tables are available as attributes on the NWBFile object.
# For display purposes, we convert the tables to pandas DataFrames to show their content.
# For a more in-depth discussion of how to access and use the tables,
# see the tutorial on :ref:`icephys_pandas_tutorial`.
pandas.set_option("display.max_columns", 6)  # avoid oversize table in the html docs
nwbfile.intracellular_recordings.to_dataframe()

#####################################################################
#

# optionally we can ignore the id columns of the category subtables
pandas.set_option("display.max_columns", 5)  # avoid oversize table in the html docs
nwbfile.intracellular_recordings.to_dataframe(ignore_category_ids=True)

#####################################################################
#
nwbfile.icephys_simultaneous_recordings.to_dataframe()

#####################################################################
#

nwbfile.icephys_sequential_recordings.to_dataframe()

#####################################################################
#

nwbfile.icephys_repetitions.to_dataframe()

#####################################################################
#

nwbfile.icephys_experimental_conditions.to_dataframe()


#####################################################################
# Validate data
# ^^^^^^^^^^^^^
#
# This section is for internal testing purposes only to validate that the roundtrip
# of the data (i.e., generate --> write --> read) produces the correct results.

# Read the data back in
with NWBHDF5IO(testpath, "r") as io:
    infile = io.read()

    # assert intracellular_recordings
    assert np.all(
        infile.intracellular_recordings.id[:] == nwbfile.intracellular_recordings.id[:]
    )

    # Assert that the ids and the VectorData, VectorIndex, and table target of the
    # recordings column of the Sweeps table are correct
    assert np.all(
        infile.icephys_simultaneous_recordings.id[:]
        == nwbfile.icephys_simultaneous_recordings.id[:]
    )
    assert np.all(
        infile.icephys_simultaneous_recordings["recordings"].target.data[:]
        == nwbfile.icephys_simultaneous_recordings["recordings"].target.data[:]
    )
    assert np.all(
        infile.icephys_simultaneous_recordings["recordings"].data[:]
        == nwbfile.icephys_simultaneous_recordings["recordings"].data[:]
    )
    assert (
        infile.icephys_simultaneous_recordings["recordings"].target.table.name
        == nwbfile.icephys_simultaneous_recordings["recordings"].target.table.name
    )

    # Assert that the ids and the VectorData, VectorIndex, and table target of the simultaneous
    #  recordings column of the SweepSequences table are correct
    assert np.all(
        infile.icephys_sequential_recordings.id[:]
        == nwbfile.icephys_sequential_recordings.id[:]
    )
    assert np.all(
        infile.icephys_sequential_recordings["simultaneous_recordings"].target.data[:]
        == nwbfile.icephys_sequential_recordings["simultaneous_recordings"].target.data[
            :
        ]
    )
    assert np.all(
        infile.icephys_sequential_recordings["simultaneous_recordings"].data[:]
        == nwbfile.icephys_sequential_recordings["simultaneous_recordings"].data[:]
    )
    assert (
        infile.icephys_sequential_recordings[
            "simultaneous_recordings"
        ].target.table.name
        == nwbfile.icephys_sequential_recordings[
            "simultaneous_recordings"
        ].target.table.name
    )

    # Assert that the ids and the VectorData, VectorIndex, and table target of the
    # sequential_recordings column of the Repetitions table are correct
    assert np.all(infile.icephys_repetitions.id[:] == nwbfile.icephys_repetitions.id[:])
    assert np.all(
        infile.icephys_repetitions["sequential_recordings"].target.data[:]
        == nwbfile.icephys_repetitions["sequential_recordings"].target.data[:]
    )
    assert np.all(
        infile.icephys_repetitions["sequential_recordings"].data[:]
        == nwbfile.icephys_repetitions["sequential_recordings"].data[:]
    )
    assert (
        infile.icephys_repetitions["sequential_recordings"].target.table.name
        == nwbfile.icephys_repetitions["sequential_recordings"].target.table.name
    )

    # Assert that the ids and the VectorData, VectorIndex, and table target of the
    # repetitions column of the Conditions table are correct
    assert np.all(
        infile.icephys_experimental_conditions.id[:]
        == nwbfile.icephys_experimental_conditions.id[:]
    )
    assert np.all(
        infile.icephys_experimental_conditions["repetitions"].target.data[:]
        == nwbfile.icephys_experimental_conditions["repetitions"].target.data[:]
    )
    assert np.all(
        infile.icephys_experimental_conditions["repetitions"].data[:]
        == nwbfile.icephys_experimental_conditions["repetitions"].data[:]
    )
    assert (
        infile.icephys_experimental_conditions["repetitions"].target.table.name
        == nwbfile.icephys_experimental_conditions["repetitions"].target.table.name
    )
    assert np.all(
        infile.icephys_experimental_conditions["tag"][:]
        == nwbfile.icephys_experimental_conditions["tag"][:]
    )



"""
.. _editing:

Editing NWB files
=================

This tutorial demonstrates how to edit NWB files in-place to make small changes to
existing containers. To add or remove containers from an NWB file, see
:ref:`modifying_data`. How and whether it is possible to edit an NWB file depends on the
storage backend and the type of edit.

.. warning::

     Manually editing an existing NWB file can make the file invalid if you are not
     careful. We highly recommend making a copy before editing and running a validation
     check on the file after editing it. See :ref:`validating`.


Editing datasets
----------------
When reading an HDF5 NWB file, PyNWB exposes :py:class:`h5py.Dataset` objects, which can
be edited in place. For this to work, you must open the file in read/write mode
(``"r+"`` or ``"a"``).

First, let's create an NWB file with data:
"""
from pynwb import NWBHDF5IO, NWBFile, TimeSeries
from datetime import datetime
from dateutil.tz import tzlocal
import numpy as np

nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier="EXAMPLE_ID",
    session_start_time=datetime.now(tzlocal()),
    session_id="LONELYMTN",
)

nwbfile.add_acquisition(
    TimeSeries(
        name="synthetic_timeseries",
        description="Random values",
        data=np.random.randn(100, 100),
        unit="m",
        rate=10e3,
    )
)

with NWBHDF5IO("test_edit.nwb", "w") as io:
    io.write(nwbfile)

##############################################
# Now, let's edit the values of the dataset

with NWBHDF5IO("test_edit.nwb", "r+") as io:
    nwbfile = io.read()
    nwbfile.acquisition["synthetic_timeseries"].data[:10] = 0.0


##############################################
# You can edit the attributes of that dataset through the ``attrs`` attribute:

with NWBHDF5IO("test_edit.nwb", "r+") as io:
    nwbfile = io.read()
    nwbfile.acquisition["synthetic_timeseries"].data.attrs["unit"] = "volts"

##############################################
# Changing the shape of dataset
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Whether it is possible to change the shape of a dataset depends on how the dataset was
# created. If the dataset was created with a flexible shape, then it is possible to
# change in-place. Creating a dataset with a flexible shape is done by specifying the
# ``maxshape`` argument of the :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO` class
# constructor. Using a ``None`` value for a component of the ``maxshape`` tuple allows
# the size of the corresponding dimension to grow, such that is can be be reset arbitrarily long
# in that dimension. Chunking is required for datasets with flexible shapes. Setting ``maxshape``,
# hence,  automatically sets chunking to ``True``, if not specified.
#
# First, let's create an NWB file with a dataset with a flexible shape:

from hdmf.backends.hdf5.h5_utils import H5DataIO

nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier="EXAMPLE_ID",
    session_start_time=datetime.now(tzlocal()),
    session_id="LONELYMTN",
)

data_io = H5DataIO(data=np.random.randn(100, 100), maxshape=(None, 100))

nwbfile.add_acquisition(
    TimeSeries(
        name="synthetic_timeseries",
        description="Random values",
        data=data_io,
        unit="m",
        rate=10e3,
    )
)

with NWBHDF5IO("test_edit2.nwb", "w") as io:
    io.write(nwbfile)

##############################################
# The ``None`` value  in the first component of ``maxshape`` means that the
# the first dimension of the dataset is unlimited. By setting the second dimension
# of ``maxshape`` to ``100``, that dimension is fixed to be no larger than ``100``.
# If you do not specify a``maxshape``, then the shape of the dataset will be fixed
# to the shape that the dataset was created with. Here, you can change the shape of
# the first dimension of this dataset.


with NWBHDF5IO("test_edit2.nwb", "r+") as io:
    nwbfile = io.read()
    nwbfile.acquisition["synthetic_timeseries"].data.resize((200, 100))

##############################################
# This will change the shape of the dataset in-place. If you try to change the shape of
# a dataset with a fixed shape, you will get an error.
#
# .. note::
#   There are several types of dataset edits that cannot be done in-place: changing the
#   shape of a dataset with a fixed shape, or changing the datatype, compression,
#   chunking, max-shape, or fill-value of a dataset. For any of these, we recommend using
#   the :py:class:`pynwb.NWBHDF5IO.export` method to export the data to a new file. See
#   :ref:`modifying_data` for more information.
#
# Editing groups
# --------------
# Editing of groups is not yet supported in PyNWB.
# To edit the attributes of a group, open the file and edit it using ``h5py``:

import h5py

with h5py.File("test_edit.nwb", "r+") as f:
    f["acquisition"]["synthetic_timeseries"].attrs["description"] = "Random values in volts"

##############################################
# .. warning::
#    Be careful not to edit values that will bring the file out of compliance with the
#    NWB specification.
#
# Renaming groups and datasets
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Rename groups and datasets in-place using the :py:meth:`~h5py.Group.move` method. For example, to rename
# the ``"synthetic_timeseries"`` group:

with h5py.File("test_edit.nwb", "r+") as f:
    f["acquisition"].move("synthetic_timeseries", "synthetic_timeseries_renamed")

##############################################
# You can use this same technique to move a group or dataset to a different location in
# the file. For example, to move the ``"synthetic_timeseries_renamed"`` group to the
# ``"analysis"`` group:

with h5py.File("test_edit.nwb", "r+") as f:
    f["acquisition"].move(
        "synthetic_timeseries_renamed",
        "/analysis/synthetic_timeseries_renamed",
    )



"""
.. _linking_data:

Modular Data Storage using External Files
===========================================

PyNWB supports linking between files using external links.

Example Use Case: Integrating data from multiple files
---------------------------------------------------------

NBWContainer classes (e.g., :py:class:`~pynwb.base.TimeSeries`) support the integration of data stored in external
HDF5 files with NWB data files via external links. To make things more concrete, let's look at the following use
case. We want to simultaneously record multiple data streams during data acquisition. Using the concept of external
links allows us to save each data stream to an external HDF5 files during data acquisition and to
afterward link the data into a single NWB file. In this case, each recording becomes represented by a
separate file-system object that can be set as read-only once the experiment is done.  In the following
we are using :py:meth:`~pynwb.base.TimeSeries` as an example, but the same approach works for other
NWBContainers as well.

.. tip::

    The same strategies we use here for creating External Links also apply to Soft Links.
    The main difference between soft and external links is that soft links point to other
    objects within the same file while external links point to objects in external files.

 .. tip::

    In the case of :py:meth:`~pynwb.base.TimeSeries`, the uncorrected timestamps generated by the acquisition
    system can be stored (or linked) in the *sync* group. In the NWB format, hardware-recorded time data
    must then be corrected to a common time base (e.g., timestamps from all hardware sources aligned) before
    it can be included in the *timestamps* of the *TimeSeries*. This means, in the case
    of :py:meth:`~pynwb.base.TimeSeries` we need to be careful that we are not including data with incompatible
    timestamps in the same file when using external links.


.. warning::

    External links can become stale/break. Since external links are pointing to data in other files
    external links may become invalid any time files are modified on the file system, e.g., renamed,
    moved or access permissions are changed.


Creating test data
^^^^^^^^^^^^^^^^^^

In the following we are creating two :py:meth:`~pynwb.base.TimeSeries` each written to a separate file.
We then show how we can integrate these files into a single NWBFile.
"""

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_linking_data.png'

from datetime import datetime
from uuid import uuid4

import numpy as np
from dateutil.tz import tzlocal

from pynwb import NWBHDF5IO, NWBFile, TimeSeries

# Create the base data
start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
data = np.arange(1000).reshape((100, 10))
timestamps = np.arange(100, dtype=float)
filename1 = "external1_example.nwb"
filename2 = "external2_example.nwb"
filename3 = "external_linkcontainer_example.nwb"
filename4 = "external_linkdataset_example.nwb"

# Create the first file
nwbfile1 = NWBFile(
    session_description="demonstrate external files",
    identifier=str(uuid4()),
    session_start_time=start_time,
)
# Create the second file
test_ts1 = TimeSeries(
    name="test_timeseries1", data=data, unit="SIunit", timestamps=timestamps
)
nwbfile1.add_acquisition(test_ts1)

# Write the first file
with NWBHDF5IO(filename1, "w") as io:
    io.write(nwbfile1)

# Create the second file
nwbfile2 = NWBFile(
    session_description="demonstrate external files",
    identifier=str(uuid4()),
    session_start_time=start_time,
)
# Create the second file
test_ts2 = TimeSeries(
    name="test_timeseries2",
    data=data,
    unit="SIunit",
    timestamps=timestamps,
)
nwbfile2.add_acquisition(test_ts2)

# Write the second file
with NWBHDF5IO(filename2, "w") as io:
    io.write(nwbfile2)


#####################
# Linking to select datasets
# ^^^^^^^^^^^^^^^^^^^^^^^^^^
#

####################
# Step 1: Create the new NWBFile
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Create the first file
nwbfile4 = NWBFile(
    session_description="demonstrate external files",
    identifier=str(uuid4()),
    session_start_time=start_time,
)


####################
# Step 2: Get the dataset you want to link to
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Now let's open our test files and retrieve our timeseries.
#

# Get the first timeseries
io1 = NWBHDF5IO(filename1, "r")
nwbfile1 = io1.read()
timeseries_1 = nwbfile1.get_acquisition("test_timeseries1")
timeseries_1_data = timeseries_1.data

####################
# Step 3: Create the object you want to link to the data
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To link to the dataset we can simply assign the data object (here `` timeseries_1.data``) to a new ``TimeSeries``

# Create a new timeseries that links to our data
test_ts4 = TimeSeries(
    name="test_timeseries4",
    data=timeseries_1_data,  # <-------
    unit="SIunit",
    timestamps=timestamps,
)
nwbfile4.add_acquisition(test_ts4)

####################
# In the above case we did not make it explicit how we want to handle the data from
# our TimeSeries, this means that :py:class:`~pynwb.NWBHDF5IO` will need to
# determine on write how to treat the dataset. We can make this explicit and customize this
# behavior on a per-dataset basis by wrapping our dataset using
# :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`

from hdmf.backends.hdf5.h5_utils import H5DataIO

# Create another timeseries that links to the same data
test_ts5 = TimeSeries(
    name="test_timeseries5",
    data=H5DataIO(data=timeseries_1_data, link_data=True),  # <-------
    unit="SIunit",
    timestamps=timestamps,
)
nwbfile4.add_acquisition(test_ts5)

####################
# Step 4: Write the data
# ~~~~~~~~~~~~~~~~~~~~~~~~
#
with NWBHDF5IO(filename4, "w") as io4:
    # Use link_data=True to specify default behavior to link rather than copy data
    io4.write(nwbfile4, link_data=True)
io1.close()

#####################
# .. note::
#
#   In the case of TimeSeries one advantage of linking to just the main dataset is that we can now
#   use our own timestamps in case the timestamps in the original file are not aligned with the
#   clock of the NWBFile we are creating. In this way we can use the linking to "re-align" different
#   TimeSeries without having to copy the main data.


####################
# Linking to whole Containers
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Appending to files and linking is made possible by passing around the same
# :py:class:`~hdmf.build.manager.BuildManager`. You can get a manager to pass around
# using the :py:meth:`~pynwb.get_manager` function.
#

from pynwb import get_manager

manager = get_manager()

####################
# .. tip::
#
#    You can pass in extensions to :py:meth:`~pynwb.get_manager` using the *extensions* argument.

####################
# Step 1: Get the container object you want to link to
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Now let's open our test files and retrieve our timeseries.
#

# Get the first timeseries
io1 = NWBHDF5IO(filename1, "r", manager=manager)
nwbfile1 = io1.read()
timeseries_1 = nwbfile1.get_acquisition("test_timeseries1")

# Get the second timeseries
io2 = NWBHDF5IO(filename2, "r", manager=manager)
nwbfile2 = io2.read()
timeseries_2 = nwbfile2.get_acquisition("test_timeseries2")

####################
# Step 2: Add the container to another NWBFile
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# To integrate both :py:meth:`~pynwb.base.TimeSeries` into a single file we simply create a new
# :py:meth:`~pynwb.file.NWBFile` and add our existing :py:meth:`~pynwb.base.TimeSeries` to it. PyNWB's
# :py:class:`~pynwb.NWBHDF5IO` backend then automatically detects that the TimeSeries have already
# been written to another file and will create external links for us.
#

# Create a new NWBFile that links to the external timeseries
nwbfile3 = NWBFile(
    session_description="demonstrate external files",
    identifier=str(uuid4()),
    session_start_time=start_time,
)
nwbfile3.add_acquisition(timeseries_1)  # <--------
nwbfile3.add_acquisition(timeseries_2)  # <--------

# Write our third file that includes our two timeseries as external links
with NWBHDF5IO(filename3, "w", manager=manager) as io3:
    io3.write(nwbfile3)
io1.close()
io2.close()


####################
# Copying an NWBFile for linking
# ------------------------------
#
# Using the :py:func:`~pynwb.file.NWBFile.copy` method allows us to easily create a shallow copy
# of a whole NWB file with links to all data in the original file. For example, we may want to
# store processed data in a new file separate from the raw data, while still being able to access
# the raw data. See the :ref:`scratch` tutorial for a detailed example.
#

####################
# Creating a single file for sharing
# -----------------------------------
#
# External links are convenient but to share data we may want to hand a single file with all the
# data to our collaborator rather than having to collect all relevant files. To do this,
# :py:class:`~hdmf.backends.hdf5.h5tools.HDF5IO` (and in turn :py:class:`~pynwb.NWBHDF5IO`)
# provide the convenience function :py:meth:`~hdmf.backends.hdf5.h5tools.HDF5IO.export`,
# which can copy the file and resolves all external links.


####################
# Automatically splitting large data across multiple HDF5 files
# -------------------------------------------------------------------
#
# For extremely large datasets it can be useful to split data across multiple files, e.g., in cases where
# the file stystem does not allow for large files. While we can achieve this by writing different
# components (e.g., :py:meth:`~pynwb.base.TimeSeries`)  to different files as described above,
# this option does not allow splitting data from single datasets. An alternative option is to use the
# ``family`` driver in ``h5py`` to automatically split the NWB file into a collection of many HDF5 files.
# The ``family`` driver stores the file on disk as a series of fixed-length chunks (each in its own file).
# In practice, to write very large arrays, we can combine this approach with :ref:`iterative_write` to
# avoid having to load all data into memory. In the example shown here we use a manual approach to
# iterative write by using :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO` to create an empty dataset and
# then filling in the data afterward.

####################
# Step 1: Create the NWBFile as usual
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

from pynwb import NWBFile
from pynwb.base import TimeSeries
from datetime import datetime
from hdmf.backends.hdf5 import H5DataIO
import numpy as np

# Create an NWBFile object
nwbfile = NWBFile(session_description='example file family',
                  identifier=str(uuid4()),
                  session_start_time=datetime.now().astimezone())

# Create the data as an empty dataset so that we can write to it later
data = H5DataIO(maxshape=(None, 10),  # make the first dimension expandable
                dtype=np.float32,     # create the data as float32
                shape=(0, 10),        # initial data shape to initialize as empty dataset
                chunks=(1000, 10)
                )

# Create a TimeSeries object
time_series = TimeSeries(name='example_timeseries',
                         data=data,
                         starting_time=0.0,
                         rate=1.0,
                         unit='mV')

# Add the TimeSeries to the NWBFile
nwbfile.add_acquisition(time_series)

####################
# Step 2: Open the new file with the `family` driver and write
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Here we need to open the file with `h5py` first to set up the driver, and then we can use
# that file with :py:class:`pynwb.NWBHDF5IO`. This is required, because :py:class:`pynwb.NWBHDF5IO`
# currently does not support passing the `memb_size` option required by the `family` driver.

import h5py
from pynwb import  NWBHDF5IO

# Define the size of the individual files, determining the number of files to create
# chunk_size = 1 * 1024**3  # 1GB per file
chunk_size = 1024**2  # 1MB just for testing

# filename pattern
filename_pattern = 'family_nwb_file_%d.nwb'

# Create the HDF5 file using the family driver
with h5py.File(name=filename_pattern, mode='w', driver='family', memb_size=chunk_size) as f:

    # Use NWBHDF5IO to write the NWBFile to the HDF5 file
    with NWBHDF5IO(file=f, mode='w') as io:
        io.write(nwbfile)

        # Write new data iteratively to the file
        for i in range(10):
            start_index = i * 1000
            stop_index = start_index + 1000
            data.dataset.resize((stop_index, 10))          # Resize the dataset
            data.dataset[start_index: stop_index , :] = i  # Set the additional values

####################
# .. note::
#
#    Alternatively, we could have also used the :ref:`iterative_write` features to write the data
#    iteratively directly as part of the `io.write` call instead of manually afterward.

####################
# Step 3: Read a file written with the family driver
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#


# Open the HDF5 file using the family driver
with h5py.File(name=filename_pattern, mode='r', driver='family', memb_size=chunk_size) as f:
    # Use NWBHDF5IO to read the NWBFile from the HDF5 file
    with NWBHDF5IO(file=f, manager=None, mode='r') as io:
        nwbfile = io.read()
        print(nwbfile)


####################
# .. note::
#
#    The filename you provide when using the ``family`` driver must contain a printf-style integer format code
#    (e.g.`%d`), which will be replaced by the file sequence number.
#
# .. note::
#
#    The ``memb_size`` parameter must be set on both write and read. As such, reading the file requires
#    the user to know the ``memb_size`` that was used for writing.
#
# .. warning::
#
#    The DANDI archive may not support NWB files that are split in this fashion.
#
# .. note::
#
#    Other file drivers, e.g., ``split`` or ``multi`` could be used in a similar fashion.
#    However, not all HDF5 drivers are supported by the the high-level API of
#    ``h5py`` and as such may require a bit more complex setup via the the
#    low-level HDF5 API in ``h5py``.
#

